# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
# stdlib
import argparse
import json
import platform
import sys
import webbrowser

# 3p
from datadog.util.format import pretty_json

# datadog
from datadog import api
from datadog.dogshell.common import report_errors, report_warnings, print_err
from datetime import datetime


class ScreenboardClient(object):
    @classmethod
    def setup_parser(cls, subparsers):
        parser = subparsers.add_parser("screenboard", help="Create, edit, and delete screenboards.")
        parser.add_argument(
            "--string_ids",
            action="store_true",
            dest="string_ids",
            help="Represent screenboard IDs as strings instead of ints in JSON",
        )

        verb_parsers = parser.add_subparsers(title="Verbs", dest="verb")
        verb_parsers.required = True

        post_parser = verb_parsers.add_parser("post", help="Create screenboards.")
        post_parser.add_argument("title", help="title for the new screenboard")
        post_parser.add_argument("description", help="short description of the screenboard")
        post_parser.add_argument(
            "graphs", help="graph definitions as a JSON string." " if unset, reads from stdin.", nargs="?"
        )
        post_parser.add_argument(
            "--template_variables",
            type=_template_variables,
            default=[],
            help="a json list of template variable dicts, e.g. "
            "[{'name': 'host', 'prefix': 'host', 'default': 'host:my-host'}]",
        )
        post_parser.add_argument("--width", type=int, default=None, help="screenboard width in pixels")
        post_parser.add_argument("--height", type=int, default=None, help="screenboard height in pixels")
        post_parser.set_defaults(func=cls._post)

        update_parser = verb_parsers.add_parser("update", help="Update existing screenboards.")
        update_parser.add_argument("screenboard_id", help="screenboard to replace " " with the new definition")
        update_parser.add_argument("title", help="title for the new screenboard")
        update_parser.add_argument("description", help="short description of the screenboard")
        update_parser.add_argument(
            "graphs", help="graph definitions as a JSON string." " if unset, reads from stdin.", nargs="?"
        )
        update_parser.add_argument(
            "--template_variables",
            type=_template_variables,
            default=[],
            help="a json list of template variable dicts, e.g. "
            "[{'name': 'host', 'prefix': 'host', 'default': "
            "'host:my-host'}]",
        )
        update_parser.add_argument("--width", type=int, default=None, help="screenboard width in pixels")
        update_parser.add_argument("--height", type=int, default=None, help="screenboard height in pixels")
        update_parser.set_defaults(func=cls._update)

        show_parser = verb_parsers.add_parser("show", help="Show a screenboard definition.")
        show_parser.add_argument("screenboard_id", help="screenboard to show")
        show_parser.set_defaults(func=cls._show)

        delete_parser = verb_parsers.add_parser("delete", help="Delete a screenboard.")
        delete_parser.add_argument("screenboard_id", help="screenboard to delete")
        delete_parser.set_defaults(func=cls._delete)

        share_parser = verb_parsers.add_parser("share", help="Share an existing screenboard's" " with a public URL.")
        share_parser.add_argument("screenboard_id", help="screenboard to share")
        share_parser.set_defaults(func=cls._share)

        revoke_parser = verb_parsers.add_parser("revoke", help="Revoke an existing screenboard's" " with a public URL.")
        revoke_parser.add_argument("screenboard_id", help="screenboard to revoke")
        revoke_parser.set_defaults(func=cls._revoke)

        pull_parser = verb_parsers.add_parser("pull", help="Pull a screenboard on the server" " into a local file")
        pull_parser.add_argument("screenboard_id", help="ID of screenboard to pull")
        pull_parser.add_argument("filename", help="file to pull screenboard into")
        pull_parser.set_defaults(func=cls._pull)

        push_parser = verb_parsers.add_parser(
            "push", help="Push updates to screenboards" " from local files to the server"
        )
        push_parser.add_argument(
            "--append_auto_text",
            action="store_true",
            dest="append_auto_text",
            help="When pushing to the server, appends filename and"
            " timestamp to the end of the screenboard description",
        )
        push_parser.add_argument(
            "file", help="screenboard files to push to the server", nargs="+", type=argparse.FileType("r")
        )
        push_parser.set_defaults(func=cls._push)

        new_file_parser = verb_parsers.add_parser(
            "new_file", help="Create a new screenboard" " and put its contents in a file"
        )
        new_file_parser.add_argument("filename", help="name of file to create with" " empty screenboard")
        new_file_parser.add_argument(
            "graphs", help="graph definitions as a JSON string." " if unset, reads from stdin.", nargs="?"
        )
        new_file_parser.set_defaults(func=cls._new_file)

    @classmethod
    def _pull(cls, args):
        cls._write_screen_to_file(args.screenboard_id, args.filename, args.timeout, args.format, args.string_ids)

    # TODO Is there a test for this one ?
    @classmethod
    def _push(cls, args):
        api._timeout = args.timeout
        for f in args.file:
            screen_obj = json.load(f)

            if args.append_auto_text:
                datetime_str = datetime.now().strftime("%x %X")
                auto_text = "<br/>\nUpdated at {0} from {1} ({2}) on {3}".format(
                    datetime_str, f.name, screen_obj["id"], platform.node()
                )
                screen_obj["description"] += auto_text

            if "id" in screen_obj:
                # Always convert to int, in case it was originally a string.
                screen_obj["id"] = int(screen_obj["id"])
                res = api.Screenboard.update(**screen_obj)
            else:
                res = api.Screenboard.create(**screen_obj)

            if "errors" in res:
                print_err("Upload of screenboard {0} from file {1} failed.".format(screen_obj["id"], f.name))

            report_warnings(res)
            report_errors(res)

            if format == "pretty":
                print(pretty_json(res))
            else:
                print(json.dumps(res))

            if args.format == "pretty":
                print("Uploaded file {0} (screenboard {1})".format(f.name, screen_obj["id"]))

    @classmethod
    def _write_screen_to_file(cls, screenboard_id, filename, timeout, format="raw", string_ids=False):
        with open(filename, "w") as f:
            res = api.Screenboard.get(screenboard_id)
            report_warnings(res)
            report_errors(res)

            screen_obj = res
            if "resource" in screen_obj:
                del screen_obj["resource"]
            if "url" in screen_obj:
                del screen_obj["url"]

            if string_ids:
                screen_obj["id"] = str(screen_obj["id"])

            json.dump(screen_obj, f, indent=2)

            if format == "pretty":
                print("Downloaded screenboard {0} to file {1}".format(screenboard_id, filename))
            else:
                print("{0} {1}".format(screenboard_id, filename))

    @classmethod
    def _post(cls, args):
        graphs = sys.stdin.read()
        api._timeout = args.timeout
        format = args.format
        graphs = args.graphs
        if args.graphs is None:
            graphs = sys.stdin.read()
        graphs = json.loads(graphs)
        res = api.Screenboard.create(
            title=args.title,
            description=args.description,
            graphs=[graphs],
            template_variables=args.template_variables,
            width=args.width,
            height=args.height,
        )
        report_warnings(res)
        report_errors(res)
        if format == "pretty":
            print(pretty_json(res))
        else:
            print(json.dumps(res))

    @classmethod
    def _update(cls, args):
        api._timeout = args.timeout
        format = args.format
        graphs = args.graphs
        if args.graphs is None:
            graphs = sys.stdin.read()
        graphs = json.loads(graphs)

        res = api.Screenboard.update(
            args.screenboard_id,
            board_title=args.title,
            description=args.description,
            widgets=graphs,
            template_variables=args.template_variables,
            width=args.width,
            height=args.height,
        )
        report_warnings(res)
        report_errors(res)
        if format == "pretty":
            print(pretty_json(res))
        else:
            print(json.dumps(res))

    @classmethod
    def _web_view(cls, args):
        dash_id = json.load(args.file)["id"]
        url = api._api_host + "/dash/dash/{0}".format(dash_id)
        webbrowser.open(url)

    @classmethod
    def _show(cls, args):
        api._timeout = args.timeout
        format = args.format
        res = api.Screenboard.get(args.screenboard_id)
        report_warnings(res)
        report_errors(res)

        if args.string_ids:
            res["id"] = str(res["id"])

        if format == "pretty":
            print(pretty_json(res))
        else:
            print(json.dumps(res))

    @classmethod
    def _delete(cls, args):
        api._timeout = args.timeout
        # TODO CHECK
        res = api.Screenboard.delete(args.screenboard_id)
        if res is not None:
            report_warnings(res)
            report_errors(res)

    @classmethod
    def _share(cls, args):
        api._timeout = args.timeout
        format = args.format
        res = api.Screenboard.share(args.screenboard_id)

        if format == "pretty":
            print(pretty_json(res))
        else:
            print(json.dumps(res))

    @classmethod
    def _revoke(cls, args):
        api._timeout = args.timeout
        format = args.format
        res = api.Screenboard.revoke(args.screenboard_id)

        if format == "pretty":
            print(pretty_json(res))
        else:
            print(json.dumps(res))

    @classmethod
    def _new_file(cls, args):
        api._timeout = args.timeout
        format = args.format
        graphs = args.graphs
        if args.graphs is None:
            graphs = sys.stdin.read()
        graphs = json.loads(graphs)
        res = api.Screenboard.create(
            board_title=args.filename, description="Description for {0}".format(args.filename), widgets=[graphs]
        )
        report_warnings(res)
        report_errors(res)

        cls._write_screen_to_file(res["id"], args.filename, args.timeout, format, args.string_ids)

        if format == "pretty":
            print(pretty_json(res))
        else:
            print(json.dumps(res))


def _template_variables(tpl_var_input):
    if "[" not in tpl_var_input:
        return [v.strip() for v in tpl_var_input.split(",")]
    else:
        try:
            return json.loads(tpl_var_input)
        except Exception:
            raise argparse.ArgumentTypeError("bad template_variable json parameter")



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
# stdlib
import os
import warnings
import sys

# 3p
import argparse

# datadog
from datadog import initialize, __version__
from datadog.dogshell.comment import CommentClient
from datadog.dogshell.common import DogshellConfig
from datadog.dogshell.dashboard_list import DashboardListClient
from datadog.dogshell.downtime import DowntimeClient
from datadog.dogshell.event import EventClient
from datadog.dogshell.host import HostClient
from datadog.dogshell.metric import MetricClient
from datadog.dogshell.monitor import MonitorClient
from datadog.dogshell.screenboard import ScreenboardClient
from datadog.dogshell.search import SearchClient
from datadog.dogshell.service_check import ServiceCheckClient
from datadog.dogshell.service_level_objective import ServiceLevelObjectiveClient
from datadog.dogshell.tag import TagClient
from datadog.dogshell.timeboard import TimeboardClient
from datadog.dogshell.dashboard import DashboardClient


def main():
    if sys.argv[0].endswith("dog"):
        warnings.warn("dog is pending deprecation. Please use dogshell instead.", PendingDeprecationWarning)

    parser = argparse.ArgumentParser(
        description="Interact with the Datadog API", formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "--config", help="location of your dogrc file (default ~/.dogrc)", default=os.path.expanduser("~/.dogrc")
    )
    parser.add_argument(
        "--api-key",
        help="your API key, from "
        "https://app.datadoghq.com/account/settings#api. "
        "You can also set the environment variables DATADOG_API_KEY or DD_API_KEY",
        dest="api_key",
        default=os.environ.get("DATADOG_API_KEY", os.environ.get("DD_API_KEY")),
    )
    parser.add_argument(
        "--application-key",
        help="your Application key, from "
        "https://app.datadoghq.com/account/settings#api. "
        "You can also set the environment variables DATADOG_APP_KEY or DD_APP_KEY",
        dest="app_key",
        default=os.environ.get("DATADOG_APP_KEY", os.environ.get("DD_APP_KEY")),
    )
    parser.add_argument(
        "--pretty",
        help="pretty-print output (suitable for human consumption, " "less useful for scripting)",
        dest="format",
        action="store_const",
        const="pretty",
    )
    parser.add_argument(
        "--raw", help="raw JSON as returned by the HTTP service", dest="format", action="store_const", const="raw"
    )
    parser.add_argument(
        "--timeout", help="time to wait in seconds before timing" " out an API call (default 10)", default=10, type=int
    )
    parser.add_argument(
        "-v", "--version", help="Dog API version", action="version", version="%(prog)s {0}".format(__version__)
    )

    config = DogshellConfig()

    # Set up subparsers for each service
    subparsers = parser.add_subparsers(title="Modes", dest="mode")
    subparsers.required = True

    CommentClient.setup_parser(subparsers)
    SearchClient.setup_parser(subparsers)
    MetricClient.setup_parser(subparsers)
    TagClient.setup_parser(subparsers)
    EventClient.setup_parser(subparsers)
    MonitorClient.setup_parser(subparsers)
    TimeboardClient.setup_parser(subparsers)
    DashboardClient.setup_parser(subparsers)
    ScreenboardClient.setup_parser(subparsers)
    DashboardListClient.setup_parser(subparsers)
    HostClient.setup_parser(subparsers)
    DowntimeClient.setup_parser(subparsers)
    ServiceCheckClient.setup_parser(subparsers)
    ServiceLevelObjectiveClient.setup_parser(subparsers)

    args = parser.parse_args()
    config.load(args.config, args.api_key, args.app_key)

    # Initialize datadog.api package
    initialize(**config)

    args.func(args)


if __name__ == "__main__":
    main()



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
from datadog.api.resources import (
    AddableAPISubResource,
    CreateableAPIResource,
    DeletableAPIResource,
    DeletableAPISubResource,
    GetableAPIResource,
    ListableAPIResource,
    ListableAPISubResource,
    UpdatableAPIResource,
    UpdatableAPISubResource,
)

from datadog.api.dashboard_list_v2 import DashboardListV2


class DashboardList(
    AddableAPISubResource,
    CreateableAPIResource,
    DeletableAPIResource,
    DeletableAPISubResource,
    GetableAPIResource,
    ListableAPIResource,
    ListableAPISubResource,
    UpdatableAPIResource,
    UpdatableAPISubResource,
):
    """
    A wrapper around Dashboard List HTTP API.
    """

    _resource_name = "dashboard/lists/manual"
    _sub_resource_name = "dashboards"

    # Support for new API version (api.DashboardList.v2)
    # Note: This needs to be removed after complete migration of these endpoints from v1 to v2.
    v2 = DashboardListV2()



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
# stdlib
import json
import logging
import time
import zlib

# datadog
from datadog.api import _api_version, _max_timeouts, _backoff_period
from datadog.api.exceptions import ClientError, ApiError, HttpBackoff, HttpTimeout, ApiNotInitialized
from datadog.api.http_client import resolve_http_client
from datadog.util.compat import is_p3k
from datadog.util.format import construct_url, construct_path, normalize_tags


log = logging.getLogger("datadog.api")


class APIClient(object):
    """
    Datadog API client: format and submit API calls to Datadog.
    Embeds a HTTP client.
    """

    # HTTP transport parameters
    _backoff_period = _backoff_period
    _max_timeouts = _max_timeouts
    _backoff_timestamp = None
    _timeout_counter = 0
    _sort_keys = False

    # Plugged HTTP client
    _http_client = None

    @classmethod
    def _get_http_client(cls):
        """
        Getter for the embedded HTTP client.
        """
        if not cls._http_client:
            cls._http_client = resolve_http_client()

        return cls._http_client

    @classmethod
    def submit(
        cls,
        method,
        path,
        api_version=None,
        body=None,
        attach_host_name=False,
        response_formatter=None,
        error_formatter=None,
        suppress_response_errors_on_codes=None,
        compress_payload=False,
        **params
    ):
        """
        Make an HTTP API request

        :param method: HTTP method to use to contact API endpoint
        :type method: HTTP method string

        :param path: API endpoint url
        :type path: url

        :param api_version: The API version used

        :param body: dictionary to be sent in the body of the request
        :type body: dictionary

        :param response_formatter: function to format JSON response from HTTP API request
        :type response_formatter: JSON input function

        :param error_formatter: function to format JSON error response from HTTP API request
        :type error_formatter: JSON input function

        :param attach_host_name: link the new resource object to the host name
        :type attach_host_name: bool

        :param suppress_response_errors_on_codes: suppress ApiError on `errors` key in the response for the given HTTP
                                                  status codes
        :type suppress_response_errors_on_codes: None|list(int)

        :param compress_payload: compress the payload using zlib
        :type compress_payload: bool

        :param params: dictionary to be sent in the query string of the request
        :type params: dictionary

        :returns: JSON or formated response from HTTP API request
        """
        try:
            # Check if it's ok to submit
            if not cls._should_submit():
                _, backoff_time_left = cls._backoff_status()
                raise HttpBackoff(backoff_time_left)

            # Import API, User and HTTP settings
            from datadog.api import (
                _api_key,
                _application_key,
                _api_host,
                _mute,
                _host_name,
                _proxies,
                _max_retries,
                _timeout,
                _cacert,
                _return_raw_response,
            )

            # Check keys and add then to params
            if _api_key is None:
                raise ApiNotInitialized("API key is not set." " Please run 'initialize' method first.")

            # Set api and app keys in headers
            headers = {}
            headers["DD-API-KEY"] = _api_key
            if _application_key:
                headers["DD-APPLICATION-KEY"] = _application_key

            # Check if the api_version is provided
            if not api_version:
                api_version = _api_version

            # set api and app keys in params only for some endpoints and thus remove keys from headers
            # as they cannot be set in both params and headers
            if cls._set_api_and_app_keys_in_params(api_version, path):
                params["api_key"] = _api_key
                del headers["DD-API-KEY"]
                if _application_key:
                    params["application_key"] = _application_key
                    del headers["DD-APPLICATION-KEY"]

            # Attach host name to body
            if attach_host_name and body:
                # Is it a 'series' list of objects ?
                if "series" in body:
                    # Adding the host name to all objects
                    for obj_params in body["series"]:
                        if obj_params.get("host", "") == "":
                            obj_params["host"] = _host_name
                else:
                    if body.get("host", "") == "":
                        body["host"] = _host_name

            # If defined, make sure tags are defined as a comma-separated string
            if "tags" in params and isinstance(params["tags"], list):
                tag_list = normalize_tags(params["tags"])
                params["tags"] = ",".join(tag_list)

            # If defined, make sure monitor_ids are defined as a comma-separated string
            if "monitor_ids" in params and isinstance(params["monitor_ids"], list):
                params["monitor_ids"] = ",".join(str(i) for i in params["monitor_ids"])

            # Process the body, if necessary
            if isinstance(body, dict):
                body = json.dumps(body, sort_keys=cls._sort_keys)
                headers["Content-Type"] = "application/json"

            if compress_payload:
                body = zlib.compress(body.encode("utf-8"))
                headers["Content-Encoding"] = "deflate"

            # Construct the URL
            url = construct_url(_api_host, api_version, path)

            # Process requesting
            start_time = time.time()

            result = cls._get_http_client().request(
                method=method,
                url=url,
                headers=headers,
                params=params,
                data=body,
                timeout=_timeout,
                max_retries=_max_retries,
                proxies=_proxies,
                verify=_cacert,
            )

            # Request succeeded: log it and reset the timeout counter
            duration = round((time.time() - start_time) * 1000.0, 4)
            log.info("%s %s %s (%sms)" % (result.status_code, method, url, duration))
            cls._timeout_counter = 0

            # Format response content
            content = result.content

            if content:
                try:
                    if is_p3k():
                        response_obj = json.loads(content.decode("utf-8"))
                    else:
                        response_obj = json.loads(content)
                except ValueError:
                    raise ValueError("Invalid JSON response: {0}".format(content))

                # response_obj can be a bool and not a dict
                if isinstance(response_obj, dict):
                    if response_obj and "errors" in response_obj:
                        # suppress ApiError when specified and just return the response
                        if not (
                            suppress_response_errors_on_codes
                            and result.status_code in suppress_response_errors_on_codes
                        ):
                            raise ApiError(response_obj)
            else:
                response_obj = None

            if response_formatter is not None:
                response_obj = response_formatter(response_obj)

            if _return_raw_response:
                return response_obj, result
            else:
                return response_obj

        except HttpTimeout:
            cls._timeout_counter += 1
            raise
        except ClientError as e:
            if _mute:
                log.error(str(e))
                if error_formatter is None:
                    return {"errors": e.args[0]}
                else:
                    return error_formatter({"errors": e.args[0]})
            else:
                raise
        except ApiError as e:
            if _mute:
                for error in e.args[0].get("errors") or []:
                    log.error(error)
                if error_formatter is None:
                    return e.args[0]
                else:
                    return error_formatter(e.args[0])
            else:
                raise

    @classmethod
    def _should_submit(cls):
        """
        Returns True if we're in a state where we should make a request
        (backoff expired, no backoff in effect), false otherwise.
        """
        now = time.time()
        should_submit = False

        # If we're not backing off, but the timeout counter exceeds the max
        # number of timeouts, then enter the backoff state, recording the time
        # we started backing off
        if not cls._backoff_timestamp and cls._timeout_counter >= cls._max_timeouts:
            log.info(
                "Max number of datadog timeouts exceeded, backing off for %s seconds",
                cls._backoff_period,
            )
            cls._backoff_timestamp = now
            should_submit = False

        # If we are backing off but the we've waiting sufficiently long enough
        # (backoff_retry_age), exit the backoff state and reset the timeout
        # counter so that we try submitting metrics again
        elif cls._backoff_timestamp:
            backed_off_time, backoff_time_left = cls._backoff_status()
            if backoff_time_left < 0:
                log.info(
                    "Exiting backoff state after %s seconds, will try to submit metrics again",
                    backed_off_time,
                )
                cls._backoff_timestamp = None
                cls._timeout_counter = 0
                should_submit = True
            else:
                log.info(
                    "In backoff state, won't submit metrics for another %s seconds",
                    backoff_time_left,
                )
                should_submit = False
        else:
            should_submit = True

        return should_submit

    @classmethod
    def _backoff_status(cls):
        """
        Get a backoff report, i.e. backoff total and remaining time.
        """
        now = time.time()
        backed_off_time = now - cls._backoff_timestamp
        backoff_time_left = cls._backoff_period - backed_off_time
        return round(backed_off_time, 2), round(backoff_time_left, 2)

    @classmethod
    def _set_api_and_app_keys_in_params(cls, api_version, path):
        """
        Some endpoints need api and app keys to be set in params only
        For these endpoints, api and app keys in headers are ignored
        :return: True if this endpoint needs api and app keys params set
        """
        constructed_path = construct_path(api_version, path)

        set_of_paths = {
            "v1/distribution_points",
            "v1/series",
            "v1/check_run",
            "v1/events",
            "v1/screen",
        }
        if constructed_path in set_of_paths:
            return True

        return False



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
"""
Datadog API resources.
"""

from datadog.api.api_client import APIClient


class CreateableAPIResource(object):
    """
    Creatable API Resource
    """

    @classmethod
    def create(cls, attach_host_name=False, method="POST", id=None, params=None, **body):
        """
        Create a new API resource object

        :param attach_host_name: link the new resource object to the host name
        :type attach_host_name: bool

        :param method: HTTP method to use to contact API endpoint
        :type method: HTTP method string

        :param id: create a new resource object as a child of the given object
        :type id: id

        :param params: new resource object source
        :type params: dictionary

        :param body: new resource object attributes
        :type body: dictionary

        :returns: Dictionary representing the API's JSON response
        """
        if params is None:
            params = {}

        path = cls._resource_name
        api_version = getattr(cls, "_api_version", None)

        if method == "GET":
            return APIClient.submit("GET", path, api_version, **body)
        if id is None:
            return APIClient.submit("POST", path, api_version, body, attach_host_name=attach_host_name, **params)

        path = "{resource_name}/{resource_id}".format(resource_name=cls._resource_name, resource_id=id)
        return APIClient.submit("POST", path, api_version, body, attach_host_name=attach_host_name, **params)


class SendableAPIResource(object):
    """
    Fork of CreateableAPIResource class with different method names
    """

    @classmethod
    def send(cls, attach_host_name=False, id=None, compress_payload=False, **body):
        """
        Create an API resource object

        :param attach_host_name: link the new resource object to the host name
        :type attach_host_name: bool

        :param id: create a new resource object as a child of the given object
        :type id: id

        :param compress_payload: compress the payload using zlib
        :type compress_payload: bool

        :param body: new resource object attributes
        :type body: dictionary

        :returns: Dictionary representing the API's JSON response
        """
        api_version = getattr(cls, "_api_version", None)

        if id is None:
            return APIClient.submit(
                "POST",
                cls._resource_name,
                api_version,
                body,
                attach_host_name=attach_host_name,
                compress_payload=compress_payload,
            )

        path = "{resource_name}/{resource_id}".format(resource_name=cls._resource_name, resource_id=id)
        return APIClient.submit(
            "POST", path, api_version, body, attach_host_name=attach_host_name, compress_payload=compress_payload
        )


class UpdatableAPIResource(object):
    """
    Updatable API Resource
    """

    @classmethod
    def update(cls, id, params=None, **body):
        """
        Update an API resource object

        :param params: updated resource object source
        :type params: dictionary

        :param body: updated resource object attributes
        :type body: dictionary

        :returns: Dictionary representing the API's JSON response
        """
        if params is None:
            params = {}

        path = "{resource_name}/{resource_id}".format(resource_name=cls._resource_name, resource_id=id)
        api_version = getattr(cls, "_api_version", None)

        return APIClient.submit("PUT", path, api_version, body, **params)


class CustomUpdatableAPIResource(object):
    """
    Updatable API Resource with custom HTTP Verb
    """

    @classmethod
    def update(cls, method=None, id=None, params=None, **body):
        """
        Update an API resource object

        :param method: HTTP method, defaults to PUT
        :type params: string

        :param params: updatable resource id
        :type params: string

        :param params: updated resource object source
        :type params: dictionary

        :param body: updated resource object attributes
        :type body: dictionary

        :returns: Dictionary representing the API's JSON response
        """

        if method is None:
            method = "PUT"
        if params is None:
            params = {}

        path = "{resource_name}/{resource_id}".format(resource_name=cls._resource_name, resource_id=id)
        api_version = getattr(cls, "_api_version", None)

        return APIClient.submit(method, path, api_version, body, **params)


class DeletableAPIResource(object):
    """
    Deletable API Resource
    """

    @classmethod
    def delete(cls, id, **params):
        """
        Delete an API resource object

        :param id: resource object to delete
        :type id: id

        :returns: Dictionary representing the API's JSON response
        """
        path = "{resource_name}/{resource_id}".format(resource_name=cls._resource_name, resource_id=id)
        api_version = getattr(cls, "_api_version", None)

        return APIClient.submit("DELETE", path, api_version, **params)


class GetableAPIResource(object):
    """
    Getable API Resource
    """

    @classmethod
    def get(cls, id, **params):
        """
        Get information about an API resource object

        :param id: resource object id to retrieve
        :type id: id

        :param params: parameters to filter API resource stream
        :type params: dictionary

        :returns: Dictionary representing the API's JSON response
        """
        path = "{resource_name}/{resource_id}".format(resource_name=cls._resource_name, resource_id=id)
        api_version = getattr(cls, "_api_version", None)

        return APIClient.submit("GET", path, api_version, **params)


class ListableAPIResource(object):
    """
    Listable API Resource
    """

    @classmethod
    def get_all(cls, **params):
        """
        List API resource objects

        :param params: parameters to filter API resource stream
        :type params: dictionary

        :returns: Dictionary representing the API's JSON response
        """
        api_version = getattr(cls, "_api_version", None)

        return APIClient.submit("GET", cls._resource_name, api_version, **params)


class ListableAPISubResource(object):
    """
    Listable API Sub-Resource
    """

    @classmethod
    def get_items(cls, id, **params):
        """
        List API sub-resource objects from a resource

        :param id: resource id to retrieve sub-resource objects from
        :type id: id

        :param params: parameters to filter API sub-resource stream
        :type params: dictionary

        :returns: Dictionary representing the API's JSON response
        """

        path = "{resource_name}/{resource_id}/{sub_resource_name}".format(
            resource_name=cls._resource_name, resource_id=id, sub_resource_name=cls._sub_resource_name
        )
        api_version = getattr(cls, "_api_version", None)

        return APIClient.submit("GET", path, api_version, **params)


class AddableAPISubResource(object):
    """
    Addable API Sub-Resource
    """

    @classmethod
    def add_items(cls, id, params=None, **body):
        """
        Add new API sub-resource objects to a resource

        :param id: resource id to add sub-resource objects to
        :type id: id

        :param params: request parameters
        :type params: dictionary

        :param body: new sub-resource objects attributes
        :type body: dictionary

        :returns: Dictionary representing the API's JSON response
        """
        if params is None:
            params = {}

        path = "{resource_name}/{resource_id}/{sub_resource_name}".format(
            resource_name=cls._resource_name, resource_id=id, sub_resource_name=cls._sub_resource_name
        )
        api_version = getattr(cls, "_api_version", None)

        return APIClient.submit("POST", path, api_version, body, **params)


class UpdatableAPISubResource(object):
    """
    Updatable API Sub-Resource
    """

    @classmethod
    def update_items(cls, id, params=None, **body):
        """
        Update API sub-resource objects of a resource

        :param id: resource id to update sub-resource objects from
        :type id: id

        :param params: request parameters
        :type params: dictionary

        :param body: updated sub-resource objects attributes
        :type body: dictionary

        :returns: Dictionary representing the API's JSON response
        """
        if params is None:
            params = {}

        path = "{resource_name}/{resource_id}/{sub_resource_name}".format(
            resource_name=cls._resource_name, resource_id=id, sub_resource_name=cls._sub_resource_name
        )
        api_version = getattr(cls, "_api_version", None)

        return APIClient.submit("PUT", path, api_version, body, **params)


class DeletableAPISubResource(object):
    """
    Deletable API Sub-Resource
    """

    @classmethod
    def delete_items(cls, id, params=None, **body):
        """
        Delete API sub-resource objects from a resource

        :param id: resource id to delete sub-resource objects from
        :type id: id

        :param params: request parameters
        :type params: dictionary

        :param body: deleted sub-resource objects attributes
        :type body: dictionary

        :returns: Dictionary representing the API's JSON response
        """
        if params is None:
            params = {}

        path = "{resource_name}/{resource_id}/{sub_resource_name}".format(
            resource_name=cls._resource_name, resource_id=id, sub_resource_name=cls._sub_resource_name
        )
        api_version = getattr(cls, "_api_version", None)

        return APIClient.submit("DELETE", path, api_version, body, **params)


class SearchableAPIResource(object):
    """
    Fork of ListableAPIResource class with different method names
    """

    @classmethod
    def _search(cls, **params):
        """
        Query an API resource stream

        :param params: parameters to filter API resource stream
        :type params: dictionary

        :returns: Dictionary representing the API's JSON response
        """
        api_version = getattr(cls, "_api_version", None)

        return APIClient.submit("GET", cls._resource_name, api_version, **params)


class ActionAPIResource(object):
    """
    Actionable API Resource
    """

    @classmethod
    def _trigger_class_action(cls, method, action_name, id=None, params=None, **body):
        """
        Trigger an action

        :param method: HTTP method to use to contact API endpoint
        :type method: HTTP method string

        :param action_name: action name
        :type action_name: string

        :param id: trigger the action for the specified resource object
        :type id: id

        :param params: action parameters
        :type params: dictionary

        :param body: action body
        :type body: dictionary

        :returns: Dictionary representing the API's JSON response
        """
        if params is None:
            params = {}

        api_version = getattr(cls, "_api_version", None)

        if id is None:
            path = "{resource_name}/{action_name}".format(resource_name=cls._resource_name, action_name=action_name)
        else:
            path = "{resource_name}/{resource_id}/{action_name}".format(
                resource_name=cls._resource_name, resource_id=id, action_name=action_name
            )
        if method == "GET":
            # Do not add body to GET requests, it causes 400 Bad request responses on EU site
            body = None
        return APIClient.submit(method, path, api_version, body, **params)

    @classmethod
    def _trigger_action(cls, method, name, id=None, **body):
        """
        Trigger an action

        :param method: HTTP method to use to contact API endpoint
        :type method: HTTP method string

        :param name: action name
        :type name: string

        :param id: trigger the action for the specified resource object
        :type id: id

        :param body: action body
        :type body: dictionary

        :returns: Dictionary representing the API's JSON response
        """
        api_version = getattr(cls, "_api_version", None)
        if id is None:
            return APIClient.submit(method, name, api_version, body)

        path = "{action_name}/{resource_id}".format(action_name=name, resource_id=id)
        if method == "GET":
            # Do not add body to GET requests, it causes 400 Bad request responses on EU site
            body = None
        return APIClient.submit(method, path, api_version, body)


class UpdatableAPISyntheticsSubResource(object):
    """
    Update Synthetics sub resource
    """

    @classmethod
    def update_synthetics_items(cls, id, params=None, **body):
        """
        Update API sub-resource objects of a resource

        :param id: resource id to update sub-resource objects from
        :type id: id

        :param params: request parameters
        :type params: dictionary

        :param body: updated sub-resource objects attributes
        :type body: dictionary

        :returns: Dictionary representing the API's JSON response
        """
        if params is None:
            params = {}

        path = "{resource_name}/tests/{resource_id}/{sub_resource_name}".format(
            resource_name=cls._resource_name, resource_id=id, sub_resource_name=cls._sub_resource_name
        )
        api_version = getattr(cls, "_api_version", None)

        return APIClient.submit("PUT", path, api_version, body, **params)


class UpdatableAPISyntheticsResource(object):
    """
    Update Synthetics resource
    """

    @classmethod
    def update_synthetics(cls, id, params=None, **body):
        """
        Update an API resource object

        :param params: updated resource object source
        :type params: dictionary

        :param body: updated resource object attributes
        :type body: dictionary

        :returns: Dictionary representing the API's JSON response
        """
        if params is None:
            params = {}

        path = "{resource_name}/tests/{resource_id}".format(resource_name=cls._resource_name, resource_id=id)
        api_version = getattr(cls, "_api_version", None)

        return APIClient.submit("PUT", path, api_version, body, **params)


class ActionAPISyntheticsResource(object):
    """
    Actionable Synthetics API Resource
    """

    @classmethod
    def _trigger_synthetics_class_action(cls, method, name, id=None, params=None, **body):
        """
        Trigger an action

        :param method: HTTP method to use to contact API endpoint
        :type method: HTTP method string

        :param name: action name
        :type name: string

        :param id: trigger the action for the specified resource object
        :type id: id

        :param params: action parameters
        :type params: dictionary

        :param body: action body
        :type body: dictionary

        :returns: Dictionary representing the API's JSON response
        """
        if params is None:
            params = {}

        api_version = getattr(cls, "_api_version", None)

        if id is None:
            path = "{resource_name}/{action_name}".format(resource_name=cls._resource_name, action_name=name)
        else:
            path = "{resource_name}/{action_name}/{resource_id}".format(
                resource_name=cls._resource_name, resource_id=id, action_name=name
            )
        if method == "GET":
            # Do not add body to GET requests, it causes 400 Bad request responses on EU site
            body = None
        return APIClient.submit(method, path, api_version, body, **params)



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
# stdlib
from collections import defaultdict

# datadog
from datadog import api
from datadog.dogshell.common import report_errors, report_warnings


class MetricClient(object):
    @classmethod
    def setup_parser(cls, subparsers):
        parser = subparsers.add_parser("metric", help="Post metrics.")
        verb_parsers = parser.add_subparsers(title="Verbs", dest="verb")
        verb_parsers.required = True

        post_parser = verb_parsers.add_parser("post", help="Post metrics")
        post_parser.add_argument("name", help="metric name")
        post_parser.add_argument("value", help="metric value (integer or decimal value)", type=float)
        post_parser.add_argument(
            "--host", help="scopes your metric to a specific host " "(default to the local host name)", default=""
        )
        post_parser.add_argument(
            "--no_host", help="no host is associated with the metric" " (overrides --host))", action="store_true"
        )
        post_parser.add_argument("--device", help="scopes your metric to a specific device", default=None)
        post_parser.add_argument("--tags", help="comma-separated list of tags", default=None)
        post_parser.add_argument(
            "--localhostname",
            help="deprecated, used to force `--host`"
            " to the local hostname "
            "(now default when no `--host` is specified)",
            action="store_true",
        )
        post_parser.add_argument(
            "--type", help="type of the metric - gauge(32bit float)" " or counter(64bit integer)", default=None
        )
        parser.set_defaults(func=cls._post)

    @classmethod
    def _post(cls, args):
        """
        Post a metric.
        """
        # Format parameters
        api._timeout = args.timeout

        host = None if args.no_host else args.host

        if args.tags:
            tags = sorted(set([t.strip() for t in args.tags.split(",") if t]))
        else:
            tags = None

        # Submit metric
        res = api.Metric.send(
            metric=args.name, points=args.value, host=host, device=args.device, tags=tags, metric_type=args.type
        )

        # Report
        res = defaultdict(list, res)

        if args.localhostname:
            # Warn about`--localhostname` command line flag deprecation
            res["warnings"].append(
                u"`--localhostname` command line flag is deprecated, made default when no `--host` "
                u"is specified. See the `--host` option for more information."
            )
        report_warnings(res)
        report_errors(res)



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
from datadog.util.compat import urlparse
from datadog.api.resources import CreateableAPIResource, ActionAPIResource, GetableAPIResource, ListableAPIResource


class Graph(CreateableAPIResource, ActionAPIResource):
    """
    A wrapper around Graph HTTP API.
    """

    _resource_name = "graph/snapshot"

    @classmethod
    def create(cls, **params):
        """
        Take a snapshot of a graph, returning the full url to the snapshot.

        :param metric_query: metric query
        :type metric_query: string query

        :param start: query start timestamp
        :type start: POSIX timestamp

        :param end: query end timestamp
        :type end: POSIX timestamp

        :param event_query: a query that will add event bands to the graph
        :type event_query: string query

        :returns: Dictionary representing the API's JSON response
        """
        return super(Graph, cls).create(method="GET", **params)

    @classmethod
    def status(cls, snapshot_url):
        """
        Returns the status code of snapshot. Can be used to know when the
        snapshot is ready for download.

        :param snapshot_url: snapshot URL to check
        :type snapshot_url: string url

        :returns: Dictionary representing the API's JSON response
        """
        snap_path = urlparse(snapshot_url).path
        snap_path = snap_path.split("/snapshot/view/")[1].split(".png")[0]

        snapshot_status_url = "graph/snapshot_status/{0}".format(snap_path)

        return super(Graph, cls)._trigger_action("GET", snapshot_status_url)


class Embed(ListableAPIResource, GetableAPIResource, ActionAPIResource, CreateableAPIResource):
    """
    A wrapper around Embed HTTP API.
    """

    _resource_name = "graph/embed"

    @classmethod
    def enable(cls, embed_id):
        """
        Enable a specified embed.

        :param embed_id: embed token
        :type embed_id: string embed token

        :returns: Dictionary representing the API's JSON response
        """
        return super(Embed, cls)._trigger_class_action("GET", id=embed_id, action_name="enable")

    @classmethod
    def revoke(cls, embed_id):
        """
        Revoke a specified embed.

        :param embed_id: embed token
        :type embed_id: string embed token

        :returns: Dictionary representing the API's JSON response
        """
        return super(Embed, cls)._trigger_class_action("GET", id=embed_id, action_name="revoke")



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc


class MetricType(object):
    Gauge = "gauge"
    Counter = "counter"
    Histogram = "histogram"
    Rate = "rate"
    Distribution = "distribution"


class MonitorType(object):
    SERVICE_CHECK = "service check"
    METRIC_ALERT = "metric alert"
    QUERY_ALERT = "query alert"
    ALL = (SERVICE_CHECK, METRIC_ALERT, QUERY_ALERT)



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
"""
A small class to run a task periodically in a thread.
"""


from threading import Thread, Event
import sys


class PeriodicTimer(Thread):
    def __init__(self, interval, function, *args, **kwargs):
        Thread.__init__(self)
        self.daemon = True
        assert interval > 0
        self.interval = interval
        assert function
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.finished = Event()

    def end(self):
        self.finished.set()

    def run(self):
        while not self.finished.wait(self.interval):
            try:
                self.function(*self.args, **self.kwargs)
            except Exception:
                # If `sys` is None, it means the interpreter is shutting down
                # and it's very likely the reason why we got an exception.
                if sys is not None:
                    raise



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
from datadog.api.resources import (
    ActionAPIResource,
    CreateableAPIResource,
    CustomUpdatableAPIResource,
    DeletableAPIResource,
    GetableAPIResource,
    ListableAPIResource,
)

from datadog.api.api_client import APIClient


class Roles(
    ActionAPIResource,
    CreateableAPIResource,
    CustomUpdatableAPIResource,
    GetableAPIResource,
    ListableAPIResource,
    DeletableAPIResource,
):
    """
    A wrapper around Tag HTTP API.
    """

    _resource_name = "roles"
    _api_version = "v2"

    @classmethod
    def update(cls, id, **body):
        """
        Update a role's attributes

        :param id: uuid of the role
        :param body: dict with type of the input, role `id`, and modified attributes
        :returns: Dictionary representing the API's JSON response
        """
        params = {}
        return super(Roles, cls).update("PATCH", id, params=params, **body)

    @classmethod
    def assign_permission(cls, id, **body):
        """
        Assign permission to a role

        :param id: uuid of the role to assign permission to
        :param body: dict with "type": "permissions" and uuid of permission to assign
        :returns: Dictionary representing the API's JSON response
        """
        params = {}
        path = "{resource_name}/{resource_id}/permissions".format(resource_name=cls._resource_name, resource_id=id)
        api_version = getattr(cls, "_api_version", None)

        return APIClient.submit("POST", path, api_version, body, **params)

    @classmethod
    def unassign_permission(cls, id, **body):
        """
        Unassign permission from a role

        :param id: uuid of the role to unassign permission from
        :param body: dict with "type": "permissions" and uuid of permission to unassign
        :returns: Dictionary representing the API's JSON response
        """
        params = {}
        path = "{resource_name}/{resource_id}/permissions".format(resource_name=cls._resource_name, resource_id=id)
        api_version = getattr(cls, "_api_version", None)

        return APIClient.submit("DELETE", path, api_version, body, **params)



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
from datadog.api.constants import CheckStatus
from datadog.api.exceptions import ApiError
from datadog.api.resources import ActionAPIResource


class ServiceCheck(ActionAPIResource):
    """
    A wrapper around ServiceCheck HTTP API.
    """

    @classmethod
    def check(cls, **body):
        """
        Post check statuses for use with monitors

        :param check: text for the message
        :type check: string

        :param host_name: name of the host submitting the check
        :type host_name: string

        :param status: integer for the status of the check
        :type status: Options: '0': OK, '1': WARNING, '2': CRITICAL, '3': UNKNOWN

        :param timestamp: timestamp of the event
        :type timestamp: POSIX timestamp

        :param message: description of why this status occurred
        :type message: string

        :param tags: list of tags for this check
        :type tags: string list

        :returns: Dictionary representing the API's JSON response
        """

        # Validate checks, include only non-null values
        for param, value in body.items():
            if param == "status" and body[param] not in CheckStatus.ALL:
                raise ApiError("Invalid status, expected one of: %s" % ", ".join(str(v) for v in CheckStatus.ALL))

        return super(ServiceCheck, cls)._trigger_action("POST", "check_run", **body)



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
"""
Metric roll-up classes.
"""
from collections import defaultdict
import random
import itertools
import threading

from datadog.util.compat import iternext
from datadog.threadstats.constants import MetricType


class Metric(object):
    """
    A base metric class that accepts points, slices them into time intervals
    and performs roll-ups within those intervals.
    """

    def add_point(self, value):
        """ Add a point to the given metric. """
        raise NotImplementedError()

    def flush(self, timestamp, interval):
        """ Flush all metrics up to the given timestamp. """
        raise NotImplementedError()


class Set(Metric):
    """ A set metric. """

    stats_tag = "g"

    def __init__(self, name, tags, host):
        self.name = name
        self.tags = tags
        self.host = host
        self.set = set()

    def add_point(self, value):
        self.set.add(value)

    def flush(self, timestamp, interval):
        return [(timestamp, len(self.set), self.name, self.tags, self.host, MetricType.Gauge, interval)]


class Gauge(Metric):
    """ A gauge metric. """

    stats_tag = "g"

    def __init__(self, name, tags, host):
        self.name = name
        self.tags = tags
        self.host = host
        self.value = None

    def add_point(self, value):
        self.value = value

    def flush(self, timestamp, interval):
        return [(timestamp, self.value, self.name, self.tags, self.host, MetricType.Gauge, interval)]


class Counter(Metric):
    """ A metric that tracks a counter value. """

    stats_tag = "c"

    def __init__(self, name, tags, host):
        self.name = name
        self.tags = tags
        self.host = host
        self.count = []

    def add_point(self, value):
        self.count.append(value)

    def flush(self, timestamp, interval):
        count = sum(self.count, 0)
        return [(timestamp, count / float(interval), self.name, self.tags, self.host, MetricType.Rate, interval)]


class Distribution(Metric):
    """ A distribution metric. """

    stats_tag = "d"

    def __init__(self, name, tags, host):
        self.name = name
        self.tags = tags
        self.host = host
        self.value = []

    def add_point(self, value):
        self.value.append(value)

    def flush(self, timestamp, interval):
        return [(timestamp, self.value, self.name, self.tags, self.host, MetricType.Distribution, interval)]


class Histogram(Metric):
    """ A histogram metric. """

    stats_tag = "h"

    def __init__(self, name, tags, host):
        self.name = name
        self.tags = tags
        self.host = host
        self.max = float("-inf")
        self.min = float("inf")
        self.sum = []
        self.iter_counter = itertools.count()
        self.count = iternext(self.iter_counter)
        self.sample_size = 1000
        self.samples = []
        self.percentiles = [0.75, 0.85, 0.95, 0.99]

    def add_point(self, value):
        self.max = self.max if self.max > value else value
        self.min = self.min if self.min < value else value
        self.sum.append(value)
        if self.count < self.sample_size:
            self.samples.append(value)
        else:
            self.samples[random.randrange(0, self.sample_size)] = value
        self.count = iternext(self.iter_counter)

    def flush(self, timestamp, interval):
        if not self.count:
            return []
        metrics = [
            (timestamp, self.min, "%s.min" % self.name, self.tags, self.host, MetricType.Gauge, interval),
            (timestamp, self.max, "%s.max" % self.name, self.tags, self.host, MetricType.Gauge, interval),
            (
                timestamp,
                self.count / float(interval),
                "%s.count" % self.name,
                self.tags,
                self.host,
                MetricType.Rate,
                interval,
            ),
            (timestamp, self.average(), "%s.avg" % self.name, self.tags, self.host, MetricType.Gauge, interval),
        ]
        length = len(self.samples)
        self.samples.sort()
        for p in self.percentiles:
            val = self.samples[int(round(p * length - 1))]
            name = "%s.%spercentile" % (self.name, int(p * 100))
            metrics.append((timestamp, val, name, self.tags, self.host, MetricType.Gauge, interval))
        return metrics

    def average(self):
        sum_metrics = sum(self.sum, 0)
        return float(sum_metrics) / self.count


class Timing(Histogram):
    """
    A timing metric.
    Inherit from Histogram to workaround and support it in API mode
    """

    stats_tag = "ms"


class MetricsAggregator(object):
    """
    A small class to handle the roll-ups of multiple metrics at once.
    """

    def __init__(self, roll_up_interval=10):
        self._lock = threading.RLock()
        self._metrics = defaultdict(lambda: {})
        self._roll_up_interval = roll_up_interval

    def add_point(self, metric, tags, timestamp, value, metric_class, sample_rate=1, host=None):
        # The sample rate is currently ignored for in process stuff
        interval = timestamp - timestamp % self._roll_up_interval
        key = (metric, host, tuple(sorted(tags)) if tags else None)
        with self._lock:
            if key not in self._metrics[interval]:
                self._metrics[interval][key] = metric_class(metric, tags, host)
            self._metrics[interval][key].add_point(value)

    def flush(self, timestamp):
        """ Flush all metrics up to the given timestamp. """
        if timestamp == float("inf"):
            interval = float("inf")
        else:
            interval = timestamp - timestamp % self._roll_up_interval

        with self._lock:
            past_intervals = [i for i in self._metrics.keys() if i < interval]
            metrics = []
            for i in past_intervals:
                for m in list(self._metrics.pop(i).values()):
                    metrics += m.flush(i, self._roll_up_interval)
        return metrics



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
# stdlib
import datetime
import time
import re
import sys
import json

# datadog
from datadog import api
from datadog.dogshell.common import report_errors, report_warnings


time_pat = re.compile(r"(?P<delta>[0-9]*\.?[0-9]+)(?P<unit>[mhd])")


def prettyprint_event(event):
    title = event["title"] or ""
    text = event.get("text", "") or ""
    handle = event.get("handle", "") or ""
    date = event["date_happened"]
    dt = datetime.datetime.fromtimestamp(date)
    link = event["url"]

    # Print
    print((title + " " + text + " " + " (" + handle + ")").strip())
    print(dt.isoformat(" ") + " | " + link)


def print_event(event):
    prettyprint_event(event)


def prettyprint_event_details(event):
    prettyprint_event(event)


def print_event_details(event):
    prettyprint_event(event)


def parse_time(timestring):
    now = time.mktime(datetime.datetime.now().timetuple())
    if timestring is None:
        t = now
    else:
        try:
            t = int(timestring)
        except Exception:
            match = time_pat.match(timestring)
            if match is None:
                raise Exception
            delta = float(match.group("delta"))
            unit = match.group("unit")
            if unit == "m":
                delta = delta * 60
            if unit == "h":
                delta = delta * 60 * 60
            if unit == "d":
                delta = delta * 60 * 60 * 24
            t = now - int(delta)
    return int(t)


class EventClient(object):
    @classmethod
    def setup_parser(cls, subparsers):
        parser = subparsers.add_parser("event", help="Post events, get event details," " and view the event stream.")
        verb_parsers = parser.add_subparsers(title="Verbs", dest="verb")
        verb_parsers.required = True

        post_parser = verb_parsers.add_parser("post", help="Post events.")
        post_parser.add_argument("title", help="event title")
        post_parser.add_argument(
            "--date_happened",
            type=int,
            help="POSIX timestamp" " when the event occurred. if unset defaults to the current time.",
        )
        post_parser.add_argument("--handle", help="user to post as. if unset, submits " "as the generic API user.")
        post_parser.add_argument("--priority", help='"normal" or "low". defaults to "normal"', default="normal")
        post_parser.add_argument(
            "--related_event_id", help="event to post as a child of." " if unset, posts a top-level event"
        )
        post_parser.add_argument("--tags", help="comma separated list of tags")
        post_parser.add_argument("--host", help="related host (default to the local host name)", default="")
        post_parser.add_argument(
            "--no_host", help="no host is associated with the event" " (overrides --host))", action="store_true"
        )
        post_parser.add_argument("--device", help="related device (e.g. eth0, /dev/sda1)")
        post_parser.add_argument("--aggregation_key", help="key to aggregate the event with")
        post_parser.add_argument("--type", help="type of event, e.g. nagios, jenkins, etc.")
        post_parser.add_argument("--alert_type", help='"error", "warning", "info" or "success". defaults to "info"')
        post_parser.add_argument("message", help="event message body. " "if unset, reads from stdin.", nargs="?")
        post_parser.set_defaults(func=cls._post)

        show_parser = verb_parsers.add_parser("show", help="Show event details.")
        show_parser.add_argument("event_id", help="event to show")
        show_parser.set_defaults(func=cls._show)

        stream_parser = verb_parsers.add_parser(
            "stream",
            help="Retrieve events from the Event Stream",
            description="Stream start and end times can be specified as either a POSIX"
            " timestamp (e.g. the output of `date +%s`) or as a period of"
            " time in the past (e.g. '5m', '6h', '3d').",
        )
        stream_parser.add_argument("start", help="start date for the stream request")
        stream_parser.add_argument("end", help="end date for the stream request " "(defaults to 'now')", nargs="?")
        stream_parser.add_argument("--priority", help="filter by priority." " 'normal' or 'low'. defaults to 'normal'")
        stream_parser.add_argument("--sources", help="comma separated list of sources to filter by")
        stream_parser.add_argument("--tags", help="comma separated list of tags to filter by")
        stream_parser.set_defaults(func=cls._stream)

    @classmethod
    def _post(cls, args):
        """
        Post an event.
        """
        api._timeout = args.timeout
        format = args.format
        message = args.message
        if message is None:
            message = sys.stdin.read()
        if args.tags is not None:
            tags = [t.strip() for t in args.tags.split(",")]
        else:
            tags = None

        host = None if args.no_host else args.host

        # Submit event
        res = api.Event.create(
            title=args.title,
            text=message,
            date_happened=args.date_happened,
            handle=args.handle,
            priority=args.priority,
            related_event_id=args.related_event_id,
            tags=tags,
            host=host,
            device=args.device,
            aggregation_key=args.aggregation_key,
            source_type_name=args.type,
            alert_type=args.alert_type,
        )

        # Report
        report_warnings(res)
        report_errors(res)
        if format == "pretty":
            prettyprint_event(res["event"])
        elif format == "raw":
            print(json.dumps(res))
        else:
            print_event(res["event"])

    @classmethod
    def _show(cls, args):
        api._timeout = args.timeout
        format = args.format
        res = api.Event.get(args.event_id)
        report_warnings(res)
        report_errors(res)
        if format == "pretty":
            prettyprint_event_details(res["event"])
        elif format == "raw":
            print(json.dumps(res))
        else:
            print_event_details(res["event"])

    @classmethod
    def _stream(cls, args):
        api._timeout = args.timeout
        format = args.format
        if args.sources is not None:
            sources = [s.strip() for s in args.sources.split(",")]
        else:
            sources = None
        if args.tags is not None:
            tags = [t.strip() for t in args.tags.split(",")]
        else:
            tags = None
        start = parse_time(args.start)
        end = parse_time(args.end)
        # res = api.Event.query(start=start, end=end)
        # TODO FIXME
        res = api.Event.query(start=start, end=end, priority=args.priority, sources=sources, tags=tags)
        report_warnings(res)
        report_errors(res)
        if format == "pretty":
            for event in res["events"]:
                prettyprint_event(event)
                print()
        elif format == "raw":
            print(json.dumps(res))
        else:
            for event in res["events"]:
                print_event(event)
                print()



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
from datadog.api.resources import (
    GetableAPIResource,
    CreateableAPIResource,
    UpdatableAPIResource,
    DeletableAPIResource,
    ActionAPIResource,
    ListableAPIResource,
)


class Screenboard(
    GetableAPIResource,
    CreateableAPIResource,
    UpdatableAPIResource,
    DeletableAPIResource,
    ActionAPIResource,
    ListableAPIResource,
):
    """
    A wrapper around Screenboard HTTP API.
    """

    _resource_name = "screen"

    @classmethod
    def share(cls, board_id):
        """
        Share the screenboard with given id

        :param board_id: screenboard to share
        :type board_id: id

        :returns: Dictionary representing the API's JSON response
        """
        return super(Screenboard, cls)._trigger_action("POST", "screen/share", board_id)

    @classmethod
    def revoke(cls, board_id):
        """
        Revoke a shared screenboard with given id

        :param board_id: screenboard to revoke
        :type board_id: id

        :returns: Dictionary representing the API's JSON response
        """
        return super(Screenboard, cls)._trigger_action("DELETE", "screen/share", board_id)



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
"""
ThreadStats is a tool for collecting application metrics without hindering
performance. It collects metrics in the application thread with very little overhead
and allows flushing metrics in process, in a thread or in a greenlet, depending
on your application's needs.
"""
import atexit
import logging
import os

# stdlib
from contextlib import contextmanager
from functools import wraps
from time import time

try:
    from time import monotonic  # type: ignore[attr-defined]
except ImportError:
    from time import time as monotonic

# datadog
from datadog.api.exceptions import ApiNotInitialized
from datadog.threadstats.constants import MetricType
from datadog.threadstats.events import EventsAggregator
from datadog.threadstats.metrics import MetricsAggregator, Counter, Gauge, Histogram, Timing, Distribution, Set
from datadog.threadstats.reporters import HttpReporter

# Loggers
log = logging.getLogger("datadog.threadstats")

DD_ENV_TAGS_MAPPING = {
    "DD_ENV": "env",
    "DD_SERVICE": "service",
    "DD_VERSION": "version",
}


class ThreadStats(object):
    def __init__(self, namespace="", constant_tags=None, compress_payload=False):
        """
        Initialize a threadstats object.

        :param namespace: Namespace to prefix all metric names
        :type namespace: string

        :param constant_tags: Tags to attach to every metric reported by this client
        :type constant_tags: list of strings

        :param compress_payload: compress the payload using zlib
        :type compress_payload: bool

        :envvar DATADOG_TAGS: Tags to attach to every metric reported by ThreadStats client
        :type DATADOG_TAGS: comma-delimited string

        :envvar DD_ENV: the env of the service running the ThreadStats client.
        If set, it is appended to the constant (global) tags of the client.
        :type DD_ENV: string

        :envvar DD_SERVICE: the name of the service running the ThreadStats client.
        If set, it is appended to the constant (global) tags of the client.
        :type DD_SERVICE: string

        :envvar DD_VERSION: the version of the service running the ThreadStats client.
        If set, it is appended to the constant (global) tags of the client.
        :type DD_VERSION: string
        """
        # Parameters
        self.namespace = namespace
        env_tags = [tag for tag in os.environ.get("DATADOG_TAGS", "").split(",") if tag]
        for var, tag_name in DD_ENV_TAGS_MAPPING.items():
            value = os.environ.get(var, "")
            if value:
                env_tags.append("{name}:{value}".format(name=tag_name, value=value))
        if constant_tags is None:
            constant_tags = []
        self.constant_tags = constant_tags + env_tags

        # State
        self._disabled = True
        self.compress_payload = compress_payload

    def start(
        self,
        flush_interval=10,
        roll_up_interval=10,
        device=None,
        flush_in_thread=True,
        flush_in_greenlet=False,
        disabled=False,
    ):
        """
        Start the ThreadStats instance with the specified metric flushing method and preferences.

        By default, metrics will be flushed in a thread.

        >>> stats.start()

        If you're running a gevent server and want to flush metrics in a
        greenlet, set *flush_in_greenlet* to True. Be sure to import and monkey
        patch gevent before starting ThreadStats. ::

        >>> from gevent import monkey; monkey.patch_all()
        >>> stats.start(flush_in_greenlet=True)

        If you'd like to flush metrics in process, set *flush_in_thread*
        to False, though you'll have to call ``flush`` manually to post metrics
        to the server. ::

        >>> stats.start(flush_in_thread=False)

        If for whatever reason, you need to disable metrics collection in a
        hurry, set ``disabled`` to True and metrics won't be collected or flushed.

        >>> stats.start(disabled=True)

        *Note:* Please remember to set your API key before,
            using datadog module ``initialize`` method.

        >>> from datadog import initialize, ThreadStats
        >>> initialize(api_key="my_api_key")
        >>> stats = ThreadStats()
        >>> stats.start()
        >>> stats.increment("home.page.hits")

        :param flush_interval: The number of seconds to wait between flushes.
        :type flush_interval: int
        :param flush_in_thread: True if you'd like to spawn a thread to flush metrics.
            It will run every `flush_interval` seconds.
        :type flush_in_thread: bool
        :param flush_in_greenlet: Set to true if you'd like to flush in a gevent greenlet.
        :type flush_in_greenlet: bool
        :param disabled: Disable metrics collection
        :type disabled: bool
        """
        self.flush_interval = flush_interval
        self.roll_up_interval = roll_up_interval
        self.device = device
        self._disabled = disabled
        self._is_auto_flushing = False

        # Create an aggregator
        self._metric_aggregator = MetricsAggregator(self.roll_up_interval)
        self._event_aggregator = EventsAggregator()

        # The reporter is responsible for sending metrics off to their final destination.
        # It's abstracted to support easy unit testing and in the near future, forwarding
        # to the datadog agent.
        self.reporter = HttpReporter(compress_payload=self.compress_payload)

        self._is_flush_in_progress = False
        self.flush_count = 0
        if self._disabled:
            log.info("ThreadStats instance is disabled. No metrics will flush.")
        else:
            if flush_in_greenlet:
                self._start_flush_greenlet()
            elif flush_in_thread:
                self._start_flush_thread()

        # Flush all remaining metrics on exit
        atexit.register(lambda: self.flush(float("inf")))

    def stop(self):
        if not self._is_auto_flushing:
            return True
        if self._flush_thread:
            self._flush_thread.end()
            self._is_auto_flushing = False
            return True

    def event(
        self,
        title,
        text,
        alert_type=None,
        aggregation_key=None,
        source_type_name=None,
        date_happened=None,
        priority=None,
        tags=None,
        hostname=None,
    ):
        """
        Send an event. Attributes are the same as the Event API. (http://docs.datadoghq.com/api/)

        >>> stats.event("Man down!", "This server needs assistance.")
        >>> stats.event("The web server restarted", \
            "The web server is up again", alert_type="success")
        """
        if not self._disabled:
            # Append all client level tags to every event
            event_tags = tags
            if self.constant_tags:
                if tags:
                    event_tags = tags + self.constant_tags
                else:
                    event_tags = self.constant_tags

            self._event_aggregator.add_event(
                title=title,
                text=text,
                alert_type=alert_type,
                aggregation_key=aggregation_key,
                source_type_name=source_type_name,
                date_happened=date_happened,
                priority=priority,
                tags=event_tags,
                host=hostname,
            )

    def gauge(self, metric_name, value, timestamp=None, tags=None, sample_rate=1, host=None):
        """
        Record the current ``value`` of a metric. The most recent value in
        a given flush interval will be recorded. Optionally, specify a set of
        tags to associate with the metric. This should be used for sum values
        such as total hard disk space, process uptime, total number of active
        users, or number of rows in a database table.

        >>> stats.gauge("process.uptime", time.time() - process_start_time)
        >>> stats.gauge("cache.bytes.free", cache.get_free_bytes(), tags=["version:1.0"])
        """
        if not self._disabled:
            self._metric_aggregator.add_point(
                metric_name, tags, timestamp or time(), value, Gauge, sample_rate=sample_rate, host=host
            )

    def set(self, metric_name, value, timestamp=None, tags=None, sample_rate=1, host=None):
        """
        Add ``value`` to the current set. The length of the set is
        flushed as a gauge to Datadog. Optionally, specify a set of
        tags to associate with the metric.

        >>> stats.set("example_metric.set", "value_1", tags=["environement:dev"])
        """
        if not self._disabled:
            self._metric_aggregator.add_point(
                metric_name, tags, timestamp or time(), value, Set, sample_rate=sample_rate, host=host
            )

    def increment(self, metric_name, value=1, timestamp=None, tags=None, sample_rate=1, host=None):
        """
        Increment the counter by the given ``value``. Optionally, specify a list of
        ``tags`` to associate with the metric. This is useful for counting things
        such as incrementing a counter each time a page is requested.

        >>> stats.increment('home.page.hits')
        >>> stats.increment('bytes.processed', file.size())
        """
        if not self._disabled:
            self._metric_aggregator.add_point(
                metric_name, tags, timestamp or time(), value, Counter, sample_rate=sample_rate, host=host
            )

    def decrement(self, metric_name, value=1, timestamp=None, tags=None, sample_rate=1, host=None):
        """
        Decrement a counter, optionally setting a value, tags and a sample
        rate.

        >>> stats.decrement("files.remaining")
        >>> stats.decrement("active.connections", 2)
        """
        if not self._disabled:
            self._metric_aggregator.add_point(
                metric_name, tags, timestamp or time(), -value, Counter, sample_rate=sample_rate, host=host
            )

    def histogram(self, metric_name, value, timestamp=None, tags=None, sample_rate=1, host=None):
        """
        Sample a histogram value. Histograms will produce metrics that
        describe the distribution of the recorded values, namely the maximum, mininum,
        average, count and the 75/85/95/99 percentiles. Optionally, specify
        a list of ``tags`` to associate with the metric.

        >>> stats.histogram("uploaded_file.size", uploaded_file.size())
        """
        if not self._disabled:
            self._metric_aggregator.add_point(
                metric_name, tags, timestamp or time(), value, Histogram, sample_rate=sample_rate, host=host
            )

    def distribution(self, metric_name, value, timestamp=None, tags=None, sample_rate=1, host=None):
        """
        Sample a distribution value. Distributions will produce metrics that
        describe the distribution of the recorded values, namely the maximum,
        median, average, count and the 50/75/90/95/99 percentiles. Optionally,
        specify a list of ``tags`` to associate with the metric.

        >>> stats.distribution("uploaded_file.size", uploaded_file.size())
        """
        if not self._disabled:
            self._metric_aggregator.add_point(
                metric_name, tags, timestamp or time(), value, Distribution, sample_rate=sample_rate, host=host
            )

    def timing(self, metric_name, value, timestamp=None, tags=None, sample_rate=1, host=None):
        """
        Record a timing, optionally setting tags and a sample rate.

        >>> stats.timing("query.response.time", 1234)
        """
        if not self._disabled:
            self._metric_aggregator.add_point(
                metric_name, tags, timestamp or time(), value, Timing, sample_rate=sample_rate, host=host
            )

    @contextmanager
    def timer(self, metric_name, sample_rate=1, tags=None, host=None):
        """
        A context manager that will track the distribution of the contained code's run time.
        Optionally specify a list of tags to associate with the metric.
        ::

            def get_user(user_id):
                with stats.timer("user.query.time"):
                    # Do what you need to ...
                    pass

            # Is equivalent to ...
            def get_user(user_id):
                start = time.time()
                try:
                    # Do what you need to ...
                    pass
                finally:
                    stats.histogram("user.query.time", time.time() - start)
        """
        start = monotonic()
        try:
            yield
        finally:
            end = monotonic()
            self.timing(metric_name, end - start, time(), tags=tags, sample_rate=sample_rate, host=host)

    def timed(self, metric_name, sample_rate=1, tags=None, host=None):
        """
        A decorator that will track the distribution of a function's run time.
        Optionally specify a list of tags to associate with the metric.
        ::

            @stats.timed("user.query.time")
            def get_user(user_id):
                # Do what you need to ...
                pass

            # Is equivalent to ...
            start = time.time()
            try:
                get_user(user_id)
            finally:
                stats.histogram("user.query.time", time.time() - start)
        """

        def wrapper(func):
            @wraps(func)
            def wrapped(*args, **kwargs):
                with self.timer(metric_name, sample_rate, tags, host):
                    result = func(*args, **kwargs)
                    return result

            return wrapped

        return wrapper

    def flush(self, timestamp=None):
        """
        Flush and post all metrics to the server. Note that this is a blocking
        call, so it is likely not suitable for user facing processes. In those
        cases, it's probably best to flush in a thread or greenlet.
        """
        try:
            if self._is_flush_in_progress:
                log.debug("A flush is already in progress. Skipping this one.")
                return False
            if self._disabled:
                log.info("Not flushing because we're disabled.")
                return False

            self._is_flush_in_progress = True

            # Process metrics
            metrics, dists = self._get_aggregate_metrics_and_dists(timestamp or time())
            count_metrics = len(metrics)
            if count_metrics:
                self.flush_count += 1
                log.debug("Flush #%s sending %s metrics" % (self.flush_count, count_metrics))
                self.reporter.flush_metrics(metrics)
            else:
                log.debug("No metrics to flush. Continuing.")

            count_dists = len(dists)
            if count_dists:
                self.flush_count += 1
                log.debug("Flush #%s sending %s distributions" % (self.flush_count, count_dists))
                self.reporter.flush_distributions(dists)
            else:
                log.debug("No distributions to flush. Continuing.")

            # Process events
            events = self._get_aggregate_events()
            count_events = len(events)
            if count_events:
                self.flush_count += 1
                log.debug("Flush #%s sending %s events" % (self.flush_count, count_events))
                self.reporter.flush_events(events)
            else:
                log.debug("No events to flush. Continuing.")
        except ApiNotInitialized:
            raise
        except Exception:
            try:
                log.exception("Error flushing metrics and events")
            except Exception:
                pass
        finally:
            self._is_flush_in_progress = False

    def _get_aggregate_metrics_and_dists(self, flush_time=None):
        """
        Get, format and return the rolled up metrics from the aggregator.
        """
        # Get rolled up metrics
        rolled_up_metrics = self._metric_aggregator.flush(flush_time)

        # FIXME: emit a dictionary from the aggregator
        metrics = []
        dists = []
        for timestamp, value, name, tags, host, metric_type, interval in rolled_up_metrics:
            metric_tags = tags
            metric_name = name

            # Append all client level tags to every metric
            if self.constant_tags:
                if tags:
                    metric_tags = tags + self.constant_tags
                else:
                    metric_tags = self.constant_tags

            # Resolve the metric name
            if self.namespace:
                metric_name = self.namespace + "." + name

            metric = {
                "metric": metric_name,
                "points": [[timestamp, value]],
                "type": metric_type,
                "host": host,
                "device": self.device,
                "tags": metric_tags,
                "interval": interval,
            }
            if metric_type == MetricType.Distribution:
                dists.append(metric)
            else:
                metrics.append(metric)
        return (metrics, dists)

    def _get_aggregate_events(self):
        # Get events
        events = self._event_aggregator.flush()
        return events

    def _start_flush_thread(self):
        """ Start a thread to flush metrics. """
        from datadog.threadstats.periodic_timer import PeriodicTimer

        if self._is_auto_flushing:
            log.info("Autoflushing already started.")
            return
        self._is_auto_flushing = True

        # A small helper for logging and flushing.
        def flush():
            try:
                log.debug("Flushing metrics in thread")
                self.flush()
            except Exception:
                try:
                    log.exception("Error flushing in thread")
                except Exception:
                    pass

        log.info("Starting flush thread with interval %s." % self.flush_interval)
        self._flush_thread = PeriodicTimer(self.flush_interval, flush)
        self._flush_thread.start()

    def _start_flush_greenlet(self):
        if self._is_auto_flushing:
            log.info("Autoflushing already started.")
            return
        self._is_auto_flushing = True

        import gevent

        # A small helper for flushing.
        def flush():
            while True:
                try:
                    log.debug("Flushing metrics in greenlet")
                    self.flush()
                    gevent.sleep(self.flush_interval)
                except Exception:
                    try:
                        log.exception("Error flushing in greenlet")
                    except Exception:
                        pass

        log.info("Starting flush greenlet with interval %s." % self.flush_interval)
        gevent.spawn(flush)



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
from datadog.api.resources import (
    ActionAPIResource,
    GetableAPIResource,
    CreateableAPIResource,
    UpdatableAPIResource,
    ListableAPIResource,
    DeletableAPIResource,
)


class User(
    ActionAPIResource,
    GetableAPIResource,
    CreateableAPIResource,
    UpdatableAPIResource,
    ListableAPIResource,
    DeletableAPIResource,
):

    _resource_name = "user"

    """
    A wrapper around User HTTP API.
    """

    @classmethod
    def invite(cls, emails):
        """
        Send an invite to join datadog to each of the email addresses in the
        *emails* list. If *emails* is a string, it will be wrapped in a list and
        sent. Returns a list of email addresses for which an email was sent.

        :param emails: emails adresses to invite to join datadog
        :type emails: string list

        :returns: Dictionary representing the API's JSON response
        """
        print("[DEPRECATION] User.invite() is deprecated. Use `create` instead.")

        if not isinstance(emails, list):
            emails = [emails]

        body = {
            "emails": emails,
        }

        return super(User, cls)._trigger_action("POST", "/invite_users", **body)



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
# datadog
from datadog.api.exceptions import ApiError
from datadog.api.format import format_points
from datadog.api.resources import SearchableAPIResource, SendableAPIResource, ListableAPIResource


class Metric(SearchableAPIResource, SendableAPIResource, ListableAPIResource):
    """
    A wrapper around Metric HTTP API
    """

    _resource_name = None

    _METRIC_QUERY_ENDPOINT = "query"
    _METRIC_SUBMIT_ENDPOINT = "series"
    _METRIC_LIST_ENDPOINT = "metrics"

    @classmethod
    def list(cls, from_epoch):
        """
        Get a list of active metrics since a given time (Unix Epoc)

        :param from_epoch: Start time in Unix Epoc (seconds)

        :returns: Dictionary containing a list of active metrics
        """

        cls._resource_name = cls._METRIC_LIST_ENDPOINT

        try:
            seconds = int(from_epoch)
            params = {"from": seconds}
        except ValueError:
            raise ApiError("Parameter 'from_epoch' must be an integer")

        return super(Metric, cls).get_all(**params)

    @staticmethod
    def _rename_metric_type(metric):
        """
        FIXME DROPME in 1.0:

        API documentation was illegitimately promoting usage of `metric_type` parameter
        instead of `type`.
        To be consistent and avoid 'backward incompatibilities', properly rename this parameter.
        """
        if "metric_type" in metric:
            metric["type"] = metric.pop("metric_type")

    @classmethod
    def send(cls, metrics=None, attach_host_name=True, compress_payload=False, **single_metric):
        """
        Submit a metric or a list of metrics to the metric API
        A metric dictionary should consist of 5 keys: metric, points, host, tags, type (some of which optional),
        see below:

        :param metric: the name of the time series
        :type metric: string

        :param compress_payload: compress the payload using zlib
        :type compress_payload: bool

        :param metrics: a list of dictionaries, each item being a metric to send
        :type metrics: list

        :param points: a (timestamp, value) pair or list of (timestamp, value) pairs
        :type points: list

        :param host: host name that produced the metric
        :type host: string

        :param tags:  list of tags associated with the metric.
        :type tags: string list

        :param type: type of the metric
        :type type: 'gauge' or 'count' or 'rate' string

        >>> api.Metric.send(metric='my.series', points=[(now, 15), (future_10s, 16)])

        >>> metrics = [{'metric': 'my.series', 'type': 'gauge', 'points': [(now, 15), (future_10s, 16)]},
                {'metric': 'my.series2', 'type': 'gauge', 'points': [(now, 15), (future_10s, 16)]}]
        >>> api.Metric.send(metrics=metrics)

        :returns: Dictionary representing the API's JSON response
        """
        # Set the right endpoint
        cls._resource_name = cls._METRIC_SUBMIT_ENDPOINT

        # Format the payload
        try:
            if metrics:
                for metric in metrics:
                    if isinstance(metric, dict):
                        cls._rename_metric_type(metric)
                        metric["points"] = format_points(metric["points"])
                metrics_dict = {"series": metrics}
            else:
                cls._rename_metric_type(single_metric)
                single_metric["points"] = format_points(single_metric["points"])
                metrics = [single_metric]
                metrics_dict = {"series": metrics}

        except KeyError:
            raise KeyError("'points' parameter is required")

        return super(Metric, cls).send(
            attach_host_name=attach_host_name, compress_payload=compress_payload, **metrics_dict
        )

    @classmethod
    def query(cls, **params):
        """
        Query metrics from Datadog

        :param start: query start timestamp
        :type start: POSIX timestamp

        :param end: query end timestamp
        :type end: POSIX timestamp

        :param query: metric query
        :type query: string query

        :returns: Dictionary representing the API's JSON response

        *start* and *end* should be less than 24 hours apart.
        It is *not* meant to retrieve metric data in bulk.

        >>> api.Metric.query(start=int(time.time()) - 3600, end=int(time.time()),
                             query='avg:system.cpu.idle{*}')
        """
        # Set the right endpoint
        cls._resource_name = cls._METRIC_QUERY_ENDPOINT

        # `from` is a reserved keyword in Python, therefore
        # `api.Metric.query(from=...)` is not permited
        # -> map `start` to `from` and `end` to `to`
        try:
            params["from"] = params.pop("start")
            params["to"] = params.pop("end")
        except KeyError as e:
            raise ApiError("The parameter '{0}' is required".format(e.args[0]))

        return super(Metric, cls)._search(**params)



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
from datadog.api.resources import ActionAPIResource, SearchableAPIResource


class Host(ActionAPIResource):
    """
    A wrapper around Host HTTP API.
    """

    _resource_name = "host"

    @classmethod
    def mute(cls, host_name, **body):
        """
        Mute a host.

        :param host_name: hostname
        :type host_name: string

        :param end: timestamp to end muting
        :type end: POSIX timestamp

        :param override: if true and the host is already muted, will override\
         existing end on the host
        :type override: bool

        :param message: message to associate with the muting of this host
        :type message: string

        :returns: Dictionary representing the API's JSON response

        """
        return super(Host, cls)._trigger_class_action("POST", "mute", host_name, **body)

    @classmethod
    def unmute(cls, host_name):
        """
        Unmute a host.

        :param host_name: hostname
        :type host_name: string

        :returns: Dictionary representing the API's JSON response

        """
        return super(Host, cls)._trigger_class_action("POST", "unmute", host_name)


class Hosts(ActionAPIResource, SearchableAPIResource):
    """
    A wrapper around Hosts HTTP API.
    """

    _resource_name = "hosts"

    @classmethod
    def search(cls, **params):
        """
        Search among hosts live within the past 2 hours. Max 100
        results at a time.

        :param filter: query to filter search results
        :type filter: string

        :param sort_field: "status", "apps", "cpu", "iowait", or "load"
        :type sort_field: string

        :param sort_dir: "asc" or "desc"
        :type sort_dir: string

        :param start: host result to start at
        :type start: integer

        :param count: number of host results to return
        :type count: integer

        :returns: Dictionary representing the API's JSOn response

        """
        return super(Hosts, cls)._search(**params)

    @classmethod
    def totals(cls):
        """
        Get total number of hosts active and up.

        :returns: Dictionary representing the API's JSON response
        """
        return super(Hosts, cls)._trigger_class_action("GET", "totals")



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
from datadog.api.resources import CreateableAPIResource, UpdatableAPIResource


class Comment(CreateableAPIResource, UpdatableAPIResource):
    """
    A wrapper around Comment HTTP API.
    """

    _resource_name = "comments"



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
from datadog.api.resources import (
    ActionAPIResource,
    CreateableAPIResource,
    CustomUpdatableAPIResource,
    DeletableAPIResource,
    GetableAPIResource,
    ListableAPIResource,
)


class Permissions(
    ActionAPIResource,
    CreateableAPIResource,
    CustomUpdatableAPIResource,
    GetableAPIResource,
    ListableAPIResource,
    DeletableAPIResource,
):
    """
    A wrapper around Tag HTTP API.
    """

    _resource_name = "permissions"
    _api_version = "v2"



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
# stdlib
import json

# 3p
from datadog.util.format import pretty_json

# datadog
from datadog import api
from datadog.dogshell.common import report_errors, report_warnings


class HostClient(object):
    @classmethod
    def setup_parser(cls, subparsers):
        parser = subparsers.add_parser("host", help="Mute, unmute hosts")
        verb_parsers = parser.add_subparsers(title="Verbs", dest="verb")
        verb_parsers.required = True

        mute_parser = verb_parsers.add_parser("mute", help="Mute a host")
        mute_parser.add_argument("host_name", help="host to mute")
        mute_parser.add_argument(
            "--end", help="POSIX timestamp, if omitted," " host will be muted until explicitly unmuted", default=None
        )
        mute_parser.add_argument("--message", help="string to associate with the" " muting of this host", default=None)
        mute_parser.add_argument(
            "--override",
            help="true/false, if true and the host is already" " muted, will overwrite existing end on the host",
            action="store_true",
        )
        mute_parser.set_defaults(func=cls._mute)

        unmute_parser = verb_parsers.add_parser("unmute", help="Unmute a host")
        unmute_parser.add_argument("host_name", help="host to mute")
        unmute_parser.set_defaults(func=cls._unmute)

    @classmethod
    def _mute(cls, args):
        api._timeout = args.timeout
        format = args.format
        res = api.Host.mute(args.host_name, end=args.end, message=args.message, override=args.override)
        report_warnings(res)
        report_errors(res)
        if format == "pretty":
            print(pretty_json(res))
        else:
            print(json.dumps(res))

    @classmethod
    def _unmute(cls, args):
        api._timeout = args.timeout
        format = args.format
        res = api.Host.unmute(args.host_name)
        report_warnings(res)
        report_errors(res)
        if format == "pretty":
            print(pretty_json(res))
        else:
            print(json.dumps(res))



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
# stdlib
import json

# 3p
from datadog.util.format import pretty_json

# datadog
from datadog import api
from datadog.dogshell.common import report_errors, report_warnings


class DashboardListClient(object):
    @classmethod
    def setup_parser(cls, subparsers):
        parser = subparsers.add_parser("dashboard_list", help="Create, edit, and delete dashboard lists")
        verb_parsers = parser.add_subparsers(title="Verbs", dest="verb")
        verb_parsers.required = True

        # Create Dashboard List parser
        post_parser = verb_parsers.add_parser("post", help="Create a dashboard list")
        post_parser.add_argument("name", help="Name for the dashboard list")
        post_parser.set_defaults(func=cls._post)

        # Update Dashboard List parser
        update_parser = verb_parsers.add_parser("update", help="Update existing dashboard list")
        update_parser.add_argument("dashboard_list_id", help="Dashboard list to replace with the new definition")
        update_parser.add_argument("name", help="Name for the dashboard list")
        update_parser.set_defaults(func=cls._update)

        # Show Dashboard List parser
        show_parser = verb_parsers.add_parser("show", help="Show a dashboard list definition")
        show_parser.add_argument("dashboard_list_id", help="Dashboard list to show")
        show_parser.set_defaults(func=cls._show)

        # Show All Dashboard Lists parser
        show_all_parser = verb_parsers.add_parser("show_all", help="Show a list of all dashboard lists")
        show_all_parser.set_defaults(func=cls._show_all)

        # Delete Dashboard List parser
        delete_parser = verb_parsers.add_parser("delete", help="Delete existing dashboard list")
        delete_parser.add_argument("dashboard_list_id", help="Dashboard list to delete")
        delete_parser.set_defaults(func=cls._delete)

        # Get Dashboards for Dashboard List parser
        get_dashboards_parser = verb_parsers.add_parser(
            "show_dashboards", help="Show a list of all dashboards for an existing dashboard list"
        )
        get_dashboards_parser.add_argument("dashboard_list_id", help="Dashboard list to show dashboards from")
        get_dashboards_parser.set_defaults(func=cls._show_dashboards)

        # Get Dashboards for Dashboard List parser (v2)
        get_dashboards_v2_parser = verb_parsers.add_parser(
            "show_dashboards_v2", help="Show a list of all dashboards for an existing dashboard list"
        )
        get_dashboards_v2_parser.add_argument("dashboard_list_id", help="Dashboard list to show dashboards from")
        get_dashboards_v2_parser.set_defaults(func=cls._show_dashboards_v2)

        # Add Dashboards to Dashboard List parser
        add_dashboards_parser = verb_parsers.add_parser(
            "add_dashboards", help="Add dashboards to an existing dashboard list"
        )
        add_dashboards_parser.add_argument("dashboard_list_id", help="Dashboard list to add dashboards to")

        add_dashboards_parser.add_argument(
            "dashboards",
            help="A JSON list of dashboard dicts, e.g. "
            + '[{"type": "custom_timeboard", "id": 1234}, '
            + '{"type": "custom_screenboard", "id": 123}]',
        )
        add_dashboards_parser.set_defaults(func=cls._add_dashboards)

        # Add Dashboards to Dashboard List parser (v2)
        add_dashboards_v2_parser = verb_parsers.add_parser(
            "add_dashboards_v2", help="Add dashboards to an existing dashboard list"
        )
        add_dashboards_v2_parser.add_argument("dashboard_list_id", help="Dashboard list to add dashboards to")
        add_dashboards_v2_parser.add_argument(
            "dashboards",
            help="A JSON list of dashboard dicts, e.g. "
            + '[{"type": "custom_timeboard", "id": "ewc-a4f-8ps"}, '
            + '{"type": "custom_screenboard", "id": "kwj-3t3-d3m"}]',
        )
        add_dashboards_v2_parser.set_defaults(func=cls._add_dashboards_v2)

        # Update Dashboards of Dashboard List parser
        update_dashboards_parser = verb_parsers.add_parser(
            "update_dashboards", help="Update dashboards of an existing dashboard list"
        )
        update_dashboards_parser.add_argument("dashboard_list_id", help="Dashboard list to update with dashboards")
        update_dashboards_parser.add_argument(
            "dashboards",
            help="A JSON list of dashboard dicts, e.g. "
            + '[{"type": "custom_timeboard", "id": 1234}, '
            + '{"type": "custom_screenboard", "id": 123}]',
        )
        update_dashboards_parser.set_defaults(func=cls._update_dashboards)

        # Update Dashboards of Dashboard List parser (v2)
        update_dashboards_v2_parser = verb_parsers.add_parser(
            "update_dashboards_v2", help="Update dashboards of an existing dashboard list"
        )
        update_dashboards_v2_parser.add_argument("dashboard_list_id", help="Dashboard list to update with dashboards")
        update_dashboards_v2_parser.add_argument(
            "dashboards",
            help="A JSON list of dashboard dicts, e.g. "
            + '[{"type": "custom_timeboard", "id": "ewc-a4f-8ps"}, '
            + '{"type": "custom_screenboard", "id": "kwj-3t3-d3m"}]',
        )
        update_dashboards_v2_parser.set_defaults(func=cls._update_dashboards_v2)

        # Delete Dashboards from Dashboard List parser
        delete_dashboards_parser = verb_parsers.add_parser(
            "delete_dashboards", help="Delete dashboards from an existing dashboard list"
        )
        delete_dashboards_parser.add_argument("dashboard_list_id", help="Dashboard list to delete dashboards from")
        delete_dashboards_parser.add_argument(
            "dashboards",
            help="A JSON list of dashboard dicts, e.g. "
            + '[{"type": "custom_timeboard", "id": 1234}, '
            + '{"type": "custom_screenboard", "id": 123}]',
        )
        delete_dashboards_parser.set_defaults(func=cls._delete_dashboards)

        # Delete Dashboards from Dashboard List parser
        delete_dashboards_v2_parser = verb_parsers.add_parser(
            "delete_dashboards", help="Delete dashboards from an existing dashboard list"
        )
        delete_dashboards_v2_parser.add_argument("dashboard_list_id", help="Dashboard list to delete dashboards from")
        delete_dashboards_v2_parser.add_argument(
            "dashboards",
            help="A JSON list of dashboard dicts, e.g. "
            + '[{"type": "custom_timeboard", "id": "ewc-a4f-8ps"}, '
            + '{"type": "custom_screenboard", "id": "kwj-3t3-d3m"}]',
        )
        delete_dashboards_v2_parser.set_defaults(func=cls._delete_dashboards_v2)

    @classmethod
    def _post(cls, args):
        api._timeout = args.timeout
        format = args.format
        name = args.name

        res = api.DashboardList.create(name=name)
        report_warnings(res)
        report_errors(res)

        if format == "pretty":
            print(pretty_json(res))
        else:
            print(json.dumps(res))

    @classmethod
    def _update(cls, args):
        api._timeout = args.timeout
        format = args.format
        dashboard_list_id = args.dashboard_list_id
        name = args.name

        res = api.DashboardList.update(dashboard_list_id, name=name)
        report_warnings(res)
        report_errors(res)

        if format == "pretty":
            print(pretty_json(res))
        else:
            print(json.dumps(res))

    @classmethod
    def _show(cls, args):
        api._timeout = args.timeout
        format = args.format
        dashboard_list_id = args.dashboard_list_id

        res = api.DashboardList.get(dashboard_list_id)
        report_warnings(res)
        report_errors(res)

        if format == "pretty":
            print(pretty_json(res))
        else:
            print(json.dumps(res))

    @classmethod
    def _show_all(cls, args):
        api._timeout = args.timeout
        format = args.format

        res = api.DashboardList.get_all()
        report_warnings(res)
        report_errors(res)

        if format == "pretty":
            print(pretty_json(res))
        else:
            print(json.dumps(res))

    @classmethod
    def _delete(cls, args):
        api._timeout = args.timeout
        format = args.format
        dashboard_list_id = args.dashboard_list_id

        res = api.DashboardList.delete(dashboard_list_id)
        report_warnings(res)
        report_errors(res)

        if format == "pretty":
            print(pretty_json(res))
        else:
            print(json.dumps(res))

    @classmethod
    def _show_dashboards(cls, args):
        api._timeout = args.timeout
        format = args.format
        dashboard_list_id = args.dashboard_list_id

        res = api.DashboardList.get_items(dashboard_list_id)
        report_warnings(res)
        report_errors(res)

        if format == "pretty":
            print(pretty_json(res))
        else:
            print(json.dumps(res))

    @classmethod
    def _show_dashboards_v2(cls, args):
        api._timeout = args.timeout
        format = args.format
        dashboard_list_id = args.dashboard_list_id

        res = api.DashboardList.v2.get_items(dashboard_list_id)
        report_warnings(res)
        report_errors(res)

        if format == "pretty":
            print(pretty_json(res))
        else:
            print(json.dumps(res))

    @classmethod
    def _add_dashboards(cls, args):
        api._timeout = args.timeout
        format = args.format
        dashboard_list_id = args.dashboard_list_id
        dashboards = json.loads(args.dashboards)

        res = api.DashboardList.add_items(dashboard_list_id, dashboards=dashboards)
        report_warnings(res)
        report_errors(res)

        if format == "pretty":
            print(pretty_json(res))
        else:
            print(json.dumps(res))

    @classmethod
    def _add_dashboards_v2(cls, args):
        api._timeout = args.timeout
        format = args.format
        dashboard_list_id = args.dashboard_list_id
        dashboards = json.loads(args.dashboards)

        res = api.DashboardList.v2.add_items(dashboard_list_id, dashboards=dashboards)
        report_warnings(res)
        report_errors(res)

        if format == "pretty":
            print(pretty_json(res))
        else:
            print(json.dumps(res))

    @classmethod
    def _update_dashboards(cls, args):
        api._timeout = args.timeout
        format = args.format
        dashboard_list_id = args.dashboard_list_id
        dashboards = json.loads(args.dashboards)

        res = api.DashboardList.update_items(dashboard_list_id, dashboards=dashboards)
        report_warnings(res)
        report_errors(res)

        if format == "pretty":
            print(pretty_json(res))
        else:
            print(json.dumps(res))

    @classmethod
    def _update_dashboards_v2(cls, args):
        api._timeout = args.timeout
        format = args.format
        dashboard_list_id = args.dashboard_list_id
        dashboards = json.loads(args.dashboards)

        res = api.DashboardList.v2.update_items(dashboard_list_id, dashboards=dashboards)
        report_warnings(res)
        report_errors(res)

        if format == "pretty":
            print(pretty_json(res))
        else:
            print(json.dumps(res))

    @classmethod
    def _delete_dashboards(cls, args):
        api._timeout = args.timeout
        format = args.format
        dashboard_list_id = args.dashboard_list_id
        dashboards = json.loads(args.dashboards)

        res = api.DashboardList.delete_items(dashboard_list_id, dashboards=dashboards)
        report_warnings(res)
        report_errors(res)

        if format == "pretty":
            print(pretty_json(res))
        else:
            print(json.dumps(res))

    @classmethod
    def _delete_dashboards_v2(cls, args):
        api._timeout = args.timeout
        format = args.format
        dashboard_list_id = args.dashboard_list_id
        dashboards = json.loads(args.dashboards)

        res = api.DashboardList.v2.delete_items(dashboard_list_id, dashboards=dashboards)
        report_warnings(res)
        report_errors(res)

        if format == "pretty":
            print(pretty_json(res))
        else:
            print(json.dumps(res))



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
from datadog.api.resources import (
    CreateableAPIResource,
    UpdatableAPIResource,
    DeletableAPIResource,
    GetableAPIResource,
    ListableAPIResource,
)


class Tag(CreateableAPIResource, UpdatableAPIResource, GetableAPIResource, ListableAPIResource, DeletableAPIResource):
    """
    A wrapper around Tag HTTP API.
    """

    _resource_name = "tags/hosts"

    @classmethod
    def create(cls, host, **body):
        """
        Add tags to a host

        :param tags: list of tags to apply to the host
        :type tags: string list

        :param source: source of the tags
        :type source: string

        :returns: Dictionary representing the API's JSON response
        """
        params = {}
        if "source" in body:
            params["source"] = body["source"]
        return super(Tag, cls).create(id=host, params=params, **body)

    @classmethod
    def update(cls, host, **body):
        """
        Update all tags for a given host

        :param tags: list of tags to apply to the host
        :type tags: string list

        :param source: source of the tags
        :type source: string

        :returns: Dictionary representing the API's JSON response
        """
        params = {}
        if "source" in body:
            params["source"] = body["source"]
        return super(Tag, cls).update(id=host, params=params, **body)



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
from datetime import datetime, timedelta
from argparse import ArgumentTypeError
import json
import re
from datadog.util.format import force_to_epoch_seconds
import time


def comma_list(list_str, item_func=None):
    if not list_str:
        raise ArgumentTypeError("Invalid comma list")
    item_func = item_func or (lambda i: i)
    return [item_func(i.strip()) for i in list_str.split(",") if i.strip()]


def comma_set(list_str, item_func=None):
    return set(comma_list(list_str, item_func=item_func))


def comma_list_or_empty(list_str):
    if not list_str:
        return []
    else:
        return comma_list(list_str)


def list_of_ints(int_csv):
    if not int_csv:
        raise ArgumentTypeError("Invalid list of ints")
    try:
        # Try as a [1, 2, 3] list
        j = json.loads(int_csv)
        if isinstance(j, (list, set)):
            j = [int(i) for i in j]
            return j
    except Exception:
        pass

    try:
        return [int(i.strip()) for i in int_csv.strip().split(",")]
    except Exception:
        raise ArgumentTypeError("Invalid list of ints: {0}".format(int_csv))


def list_of_ints_and_strs(csv):
    def int_or_str(item):
        try:
            return int(item)
        except ValueError:
            return item

    return comma_list(csv, int_or_str)


def set_of_ints(int_csv):
    return set(list_of_ints(int_csv))


class DateParsingError(Exception):
    """Thrown if parse_date exhausts all possible parsings of a string"""


_date_fieldre = re.compile(r"(\d+)\s?(\w+) (ago|ahead)")


def _midnight():
    """ Truncate a date to midnight. Default to UTC midnight today."""
    return datetime.utcnow().replace(hour=0, minute=0, second=0, microsecond=0)


def parse_date_as_epoch_timestamp(date_str):
    return parse_date(date_str, to_epoch_ts=True)


def _parse_date_noop_formatter(d):
    """ NOOP - only here for pylint """
    return d


def parse_date(date_str, to_epoch_ts=False):
    formatter = _parse_date_noop_formatter
    if to_epoch_ts:
        formatter = force_to_epoch_seconds

    if isinstance(date_str, datetime):
        return formatter(date_str)
    elif isinstance(date_str, time.struct_time):
        return formatter(datetime.fromtimestamp(time.mktime(date_str)))

    # Parse relative dates.
    if date_str == "today":
        return formatter(_midnight())
    elif date_str == "yesterday":
        return formatter(_midnight() - timedelta(days=1))
    elif date_str == "tomorrow":
        return formatter(_midnight() + timedelta(days=1))
    elif date_str.endswith(("ago", "ahead")):
        m = _date_fieldre.match(date_str)
        if m:
            fields = m.groups()
        else:
            fields = date_str.split(" ")[1:]
        num = int(fields[0])
        short_unit = fields[1]
        time_direction = {"ago": -1, "ahead": 1}[fields[2]]
        assert short_unit, short_unit
        units = ["weeks", "days", "hours", "minutes", "seconds"]
        # translate 'h' -> 'hours'
        short_units = dict([(u[:1], u) for u in units])
        unit = short_units.get(short_unit, short_unit)
        # translate 'hour' -> 'hours'
        if unit[-1] != "s":
            unit += "s"  # tolerate 1 hour
        assert unit in units, "'%s' not in %s" % (unit, units)
        return formatter(datetime.utcnow() + time_direction * timedelta(**{unit: num}))
    elif date_str == "now":
        return formatter(datetime.utcnow())

    def _from_epoch_timestamp(seconds):
        print("_from_epoch_timestamp({})".format(seconds))
        return datetime.utcfromtimestamp(float(seconds))

    def _from_epoch_ms_timestamp(millis):
        print("_from_epoch_ms_timestamp({})".format(millis))
        in_sec = float(millis) / 1000.0
        print("_from_epoch_ms_timestamp({}) -> {}".format(millis, in_sec))
        return _from_epoch_timestamp(in_sec)

    # Or parse date formats (most specific to least specific)
    parse_funcs = [
        lambda d: datetime.strptime(d, "%Y-%m-%d %H:%M:%S.%f"),
        lambda d: datetime.strptime(d, "%Y-%m-%d %H:%M:%S"),
        lambda d: datetime.strptime(d, "%Y-%m-%dT%H:%M:%S.%f"),
        lambda d: datetime.strptime(d, "%Y-%m-%dT%H:%M:%S"),
        lambda d: datetime.strptime(d, "%Y-%m-%d %H:%M"),
        lambda d: datetime.strptime(d, "%Y-%m-%d-%H"),
        lambda d: datetime.strptime(d, "%Y-%m-%d"),
        lambda d: datetime.strptime(d, "%Y-%m"),
        lambda d: datetime.strptime(d, "%Y"),
        _from_epoch_timestamp,  # an epoch in seconds
        _from_epoch_ms_timestamp,  # an epoch in milliseconds
    ]

    for parse_func in parse_funcs:
        try:
            return formatter(parse_func(date_str))
        except Exception:
            pass
    raise DateParsingError(u"Could not parse {0} as date".format(date_str))



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
from datadog.util.format import force_to_epoch_seconds
from datadog.api.resources import (
    GetableAPIResource,
    CreateableAPIResource,
    UpdatableAPIResource,
    ListableAPIResource,
    DeletableAPIResource,
    ActionAPIResource,
)


class ServiceLevelObjective(
    GetableAPIResource,
    CreateableAPIResource,
    UpdatableAPIResource,
    ListableAPIResource,
    DeletableAPIResource,
    ActionAPIResource,
):
    """
    A wrapper around Service Level Objective HTTP API.
    """

    _resource_name = "slo"

    @classmethod
    def create(cls, attach_host_name=False, method="POST", id=None, params=None, **body):
        """
        Create a SLO

        :returns: created SLO details
        """
        return super(ServiceLevelObjective, cls).create(
            attach_host_name=False, method="POST", id=None, params=params, **body
        )

    @classmethod
    def get(cls, id, **params):
        """
        Get a specific SLO details.

        :param id: SLO id to get details for
        :type id: str

        :returns: SLO details
        """
        return super(ServiceLevelObjective, cls).get(id, **params)

    @classmethod
    def get_all(cls, query=None, ids=None, offset=0, limit=100, **params):
        """
        Get all SLO details.

        :param query: optional search query - syntax in UI && online documentation
        :type query: str

        :param ids: optional list of SLO ids to get many specific SLOs at once.
        :type ids: list(str)

        :param offset: offset of results to use (default 0)
        :type offset: int

        :param limit: limit of results to return (default: 100)
        :type limit: int

        :returns: SLOs matching the query
        """
        search_terms = {}
        if query:
            search_terms["query"] = query
        if ids:
            search_terms["ids"] = ids
        search_terms["offset"] = offset
        search_terms["limit"] = limit

        return super(ServiceLevelObjective, cls).get_all(**search_terms)

    @classmethod
    def update(cls, id, params=None, **body):
        """
        Update a specific SLO details.

        :param id: SLO id to update details for
        :type id: str

        :returns: SLO details
        """
        return super(ServiceLevelObjective, cls).update(id, params, **body)

    @classmethod
    def delete(cls, id, **params):
        """
        Delete a specific SLO.

        :param id: SLO id to delete
        :type id: str

        :returns: SLO ids removed
        """
        return super(ServiceLevelObjective, cls).delete(id, **params)

    @classmethod
    def bulk_delete(cls, ops, **params):
        """
        Bulk Delete Timeframes from multiple SLOs.

        :param ops: a dictionary mapping of SLO ID to timeframes to remove.
        :type ops: dict(str, list(str))

        :returns: Dictionary representing the API's JSON response
            `errors` - errors with operation
            `data` - updates and deletions
        """
        return super(ServiceLevelObjective, cls)._trigger_class_action(
            "POST",
            "bulk_delete",
            body=ops,
            params=params,
            suppress_response_errors_on_codes=[200],
        )

    @classmethod
    def delete_many(cls, ids, **params):
        """
        Delete Multiple SLOs

        :param ids: a list of SLO IDs to remove
        :type ids: list(str)

        :returns: Dictionary representing the API's JSON response see `data` list(slo ids) && `errors`
        """
        return super(ServiceLevelObjective, cls)._trigger_class_action(
            "DELETE",
            "",
            params=params,
            body={"ids": ids},
            suppress_response_errors_on_codes=[200],
        )

    @classmethod
    def can_delete(cls, ids, **params):
        """
        Check if the following SLOs can be safely deleted.

        This is used to check if SLO has any references to it.

        :param ids: a list of SLO IDs to check
        :type ids: list(str)

        :returns: Dictionary representing the API's JSON response
                  "data.ok" represents a list of SLO ids that have no known references.
                  "errors" contains a dictionary of SLO ID to known reference(s).
        """
        params["ids"] = ids
        return super(ServiceLevelObjective, cls)._trigger_class_action(
            "GET",
            "can_delete",
            params=params,
            body=None,
            suppress_response_errors_on_codes=[200],
        )

    @classmethod
    def history(cls, id, from_ts, to_ts, **params):
        """
        Get the SLO's history from the given time range.

        :param id: SLO ID to query
        :type id: str

        :param from_ts: `from` timestamp in epoch seconds to query
        :type from_ts: int|datetime.datetime

        :param to_ts: `to` timestamp in epoch seconds to query, must be > `from_ts`
        :type to_ts: int|datetime.datetime

        :returns: Dictionary representing the API's JSON response
                  "data.ok" represents a list of SLO ids that have no known references.
                  "errors" contains a dictionary of SLO ID to known reference(s).
        """
        params["id"] = id
        params["from_ts"] = force_to_epoch_seconds(from_ts)
        params["to_ts"] = force_to_epoch_seconds(to_ts)
        return super(ServiceLevelObjective, cls)._trigger_class_action(
            "GET",
            "history",
            id=id,
            params=params,
            body=None,
            suppress_response_errors_on_codes=[200],
        )



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
# stdlib
import calendar
import datetime
import json
import re

TAG_INVALID_CHARS_RE = re.compile(r"[^\w\d_\-:/\.]", re.UNICODE)
TAG_INVALID_CHARS_SUBS = "_"


def pretty_json(obj):
    return json.dumps(obj, sort_keys=True, indent=2)


def construct_url(host, api_version, path):
    return "{}/api/{}/{}".format(host.strip("/"), api_version.strip("/"), path.strip("/"))


def construct_path(api_version, path):
    return "{}/{}".format(api_version.strip("/"), path.strip("/"))


def force_to_epoch_seconds(epoch_sec_or_dt):
    if isinstance(epoch_sec_or_dt, datetime.datetime):
        return calendar.timegm(epoch_sec_or_dt.timetuple())
    return epoch_sec_or_dt


def normalize_tags(tag_list):
    return [TAG_INVALID_CHARS_RE.sub(TAG_INVALID_CHARS_SUBS, tag) for tag in tag_list]



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
from datadog.api.resources import (
    GetableAPIResource,
    CreateableAPIResource,
    UpdatableAPIResource,
    ListableAPIResource,
    DeletableAPIResource,
)


class Timeboard(
    GetableAPIResource, CreateableAPIResource, UpdatableAPIResource, ListableAPIResource, DeletableAPIResource
):
    """
    A wrapper around Timeboard HTTP API.
    """

    _resource_name = "dash"



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
# stdlib
import json

# datadog
from datadog import api
from datadog.dogshell.common import report_errors, report_warnings


class TagClient(object):
    @classmethod
    def setup_parser(cls, subparsers):
        parser = subparsers.add_parser("tag", help="View and modify host tags.")
        verb_parsers = parser.add_subparsers(title="Verbs", dest="verb")
        verb_parsers.required = True

        add_parser = verb_parsers.add_parser(
            "add", help="Add a host to one or more tags.", description="Hosts can be specified by name or id."
        )
        add_parser.add_argument("host", help="host to add")
        add_parser.add_argument("tag", help="tag to add host to (one or more, space separated)", nargs="+")
        add_parser.set_defaults(func=cls._add)

        replace_parser = verb_parsers.add_parser(
            "replace",
            help="Replace all tags with one or more new tags.",
            description="Hosts can be specified by name or id.",
        )
        replace_parser.add_argument("host", help="host to modify")
        replace_parser.add_argument("tag", help="list of tags to add host to", nargs="+")
        replace_parser.set_defaults(func=cls._replace)

        show_parser = verb_parsers.add_parser(
            "show", help="Show host tags.", description="Hosts can be specified by name or id."
        )
        show_parser.add_argument("host", help="host to show (or 'all' to show all tags)")
        show_parser.set_defaults(func=cls._show)

        detach_parser = verb_parsers.add_parser(
            "detach", help="Remove a host from all tags.", description="Hosts can be specified by name or id."
        )
        detach_parser.add_argument("host", help="host to detach")
        detach_parser.set_defaults(func=cls._detach)

    @classmethod
    def _add(cls, args):
        api._timeout = args.timeout
        format = args.format
        res = api.Tag.create(args.host, tags=args.tag)
        report_warnings(res)
        report_errors(res)
        if format == "pretty":
            print("Tags for '%s':" % res["host"])
            for c in res["tags"]:
                print("  " + c)
        elif format == "raw":
            print(json.dumps(res))
        else:
            for c in res["tags"]:
                print(c)

    @classmethod
    def _replace(cls, args):
        api._timeout = args.timeout
        format = args.format
        res = api.Tag.update(args.host, tags=args.tag)
        report_warnings(res)
        report_errors(res)
        if format == "pretty":
            print("Tags for '%s':" % res["host"])
            for c in res["tags"]:
                print("  " + c)
        elif format == "raw":
            print(json.dumps(res))
        else:
            for c in res["tags"]:
                print(c)

    @classmethod
    def _show(cls, args):
        api._timeout = args.timeout
        format = args.format
        if args.host == "all":
            res = api.Tag.get_all()
        else:
            res = api.Tag.get(args.host)
        report_warnings(res)
        report_errors(res)
        if args.host == "all":
            if format == "pretty":
                for tag, hosts in list(res["tags"].items()):
                    for host in hosts:
                        print(tag)
                        print("  " + host)
                    print()
            elif format == "raw":
                print(json.dumps(res))
            else:
                for tag, hosts in list(res["tags"].items()):
                    for host in hosts:
                        print(tag + "\t" + host)
        else:
            if format == "pretty":
                for tag in res["tags"]:
                    print(tag)
            elif format == "raw":
                print(json.dumps(res))
            else:
                for tag in res["tags"]:
                    print(tag)

    @classmethod
    def _detach(cls, args):
        api._timeout = args.timeout
        res = api.Tag.delete(args.host)
        if res is not None:
            report_warnings(res)
            report_errors(res)



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
from datadog.api.resources import (
    GetableAPIResource,
    CreateableAPIResource,
    DeletableAPIResource,
    UpdatableAPIResource,
    UpdatableAPISubResource,
    ListableAPISubResource,
)


class AwsIntegration(
    GetableAPIResource,
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPISubResource,
    UpdatableAPIResource,
    UpdatableAPISubResource,
):
    """
    A wrapper around AWS Integration API.
    """

    _resource_name = "integration"
    _resource_id = "aws"

    @classmethod
    def list(cls, **params):
        """
        List all Datadog-AWS integrations available in your Datadog organization.

        >>> api.AwsIntegration.list()
        """
        return super(AwsIntegration, cls).get(id=cls._resource_id, **params)

    @classmethod
    def create(cls, **params):
        """
        Add a new AWS integration config.

        :param account_id: Your AWS Account ID without dashes. \
        Consult the Datadog AWS integration to learn more about \
        your AWS account ID.
        :type account_id: string

        :param access_key_id: If your AWS account is a GovCloud \
        or China account, enter the corresponding Access Key ID.
        :type access_key_id: string

        :param role_name: Your Datadog role delegation name. \
        For more information about you AWS account Role name, \
        see the Datadog AWS integration configuration info.
        :type role_name: string

        :param filter_tags: The array of EC2 tags (in the form key:value) \
        defines a filter that Datadog uses when collecting metrics from EC2. \
        Wildcards, such as ? (for single characters) and * (for multiple characters) \
        can also be used. Only hosts that match one of the defined tags will be imported \
        into Datadog. The rest will be ignored. Host matching a given tag can also be \
        excluded by adding ! before the tag. e.x. \
        env:production,instance-type:c1.*,!region:us-east-1 For more information \
        on EC2 tagging, see the AWS tagging documentation.
        :type filter_tags: list of strings

        :param host_tags: Array of tags (in the form key:value) to add to all hosts and \
        metrics reporting through this integration.
        :type host_tags: list of strings

        :param account_specific_namespace_rules: An object (in the form \
        {"namespace1":true/false, "namespace2":true/false}) that enables \
        or disables metric collection for specific AWS namespaces for this \
        AWS account only. A list of namespaces can be found at the \
        /v1/integration/aws/available_namespace_rules endpoint.
        :type account_specific_namespace_rules: dictionary

        :param excluded_regions: An array of AWS regions to exclude \
        from metrics collection.
        :type excluded_regions: list of strings

        :returns: Dictionary representing the API's JSON response

        >>> account_id = "<AWS_ACCOUNT_ID>"
        >>> access_key_id = "<AWS_ACCESS_KEY_ID>"
        >>> role_name = "DatadogAwsRole"
        >>> filter_tags = ["<KEY>:<VALUE>"]
        >>> host_tags = ["<KEY>:<VALUE>"]
        >>> account_specific_namespace_rules = {"namespace1":true/false, "namespace2":true/false}
        >>> excluded_regions = ["us-east-1", "us-west-1"]

        >>> api.AwsIntegration.create(account_id=account_id, role_name=role_name, \
        filter_tags=filter_tags,host_tags=host_tags,\
        account_specific_namespace_rules=account_specific_namespace_rules \
        excluded_regions=excluded_regions)
        """
        return super(AwsIntegration, cls).create(id=cls._resource_id, **params)

    @classmethod
    def update(cls, **body):
        """
        Update an AWS integration config.

        :param account_id: Your existing AWS Account ID without dashes. \
        Consult the Datadog AWS integration to learn more about \
        your AWS account ID.
        :type account_id: string

        :param new_account_id: Your new AWS Account ID without dashes. \
        Consult the Datadog AWS integration to learn more about \
        your AWS account ID. This is the account to be updated.
        :type new_account_id: string

        :param role_name: Your existing Datadog role delegation name. \
        For more information about you AWS account Role name, \
        see the Datadog AWS integration configuration info.
        :type role_name: string

        :param new_role_name: Your new Datadog role delegation name. \
        For more information about you AWS account Role name, \
        see the Datadog AWS integration configuration info. \
        This is the role_name to be updated.
        :type new_role_name: string

        :param access_key_id: If your AWS account is a GovCloud \
        or China account, enter the existing Access Key ID.
        :type access_key_id: string

        :param new_access_key_id: If your AWS account is a GovCloud \
        or China account, enter the new Access Key ID to be set.
        :type new_access_key_id: string

        :param secret_access_key: If your AWS account is a GovCloud \
        or China account, enter the existing Secret Access Key.
        :type secret_access_key: string

        :param new_secret_access_key: If your AWS account is a GovCloud \
        or China account, enter the new key to be set.
        :type new_secret_access_key: string

        :param filter_tags: The array of EC2 tags (in the form key:value) \
        defines a filter that Datadog uses when collecting metrics from EC2. \
        Wildcards, such as ? (for single characters) and * (for multiple characters) \
        can also be used. Only hosts that match one of the defined tags will be imported \
        into Datadog. The rest will be ignored. Host matching a given tag can also be \
        excluded by adding ! before the tag. e.x. \
        env:production,instance-type:c1.*,!region:us-east-1 For more information \
        on EC2 tagging, see the AWS tagging documentation.
        :type filter_tags: list of strings

        :param host_tags: Array of tags (in the form key:value) to add to all hosts and \
        metrics reporting through this integration.
        :type host_tags: list of strings

        :param account_specific_namespace_rules: An object (in the form \
        {"namespace1":true/false, "namespace2":true/false}) that enables \
        or disables metric collection for specific AWS namespaces for this \
        AWS account only. A list of namespaces can be found at the \
        /v1/integration/aws/available_namespace_rules endpoint.
        :type account_specific_namespace_rules: dictionary

        :param excluded_regions: An array of AWS regions to exclude \
        from metrics collection.
        :type excluded_regions: list of strings

        :returns: Dictionary representing the API's JSON response

        The following will depend on whether role delegation or access keys are being used.
        If using role delegation, use the fields for role_name and account_id.
        For access keys, use fields for access_key_id and secret_access_key.

        Both the existing fields and new fields are required no matter what. i.e. If the config is \
        account_id/role_name based, then `account_id`, `role_name`, `new_account_id`, and \
        `new_role_name` are all required.

        For access_key based accounts, `access_key_id`, `secret_access_key`, `new_access_key_id`, \
        and `new_secret_access_key` are all required.

        >>> account_id = "<EXISTING_AWS_ACCOUNT_ID>"
        >>> role_name = "<EXISTING_AWS_ROLE_NAME>"
        >>> access_key_id = "<EXISTING_AWS_ACCESS_KEY_ID>"
        >>> secret_access_key = "<EXISTING_AWS_SECRET_ACCESS_KEY>"
        >>> new_account_id = "<NEW_AWS_ACCOUNT_ID>"
        >>> new_role_name = "<NEW_AWS_ROLE_NAME>"
        >>> new_access_key_id = "<NEW_AWS_ACCESS_KEY_ID>"
        >>> new_secret_access_key = "<NEW_AWS_SECRET_ACCESS_KEY_ID>"
        >>> filter_tags = ["<KEY>:<VALUE>"]
        >>> host_tags = ["<KEY>:<VALUE>"]
        >>> account_specific_namespace_rules = {"namespace1":true/false, "namespace2":true/false}
        >>> excluded_regions = ["us-east-1", "us-west-1"]

        >>> api.AwsIntegration.update(account_id=account_id, role_name=role_name, \
        new_account_id=new_account_id, new_role_name=new_role_name, \
        filter_tags=filter_tags,host_tags=host_tags,\
        account_specific_namespace_rules=account_specific_namespace_rules, \
        excluded_regions=excluded_regions)
        """
        params = {}
        if body.get("account_id") and body.get("role_name"):
            params["account_id"] = body.pop("account_id")
            params["role_name"] = body.pop("role_name")
            if body.get("new_account_id"):
                body["account_id"] = body.pop("new_account_id")
            if body.get("new_role_name"):
                body["role_name"] = body.pop("new_role_name")
        if body.get("access_key_id") and body.get("secret_access_key"):
            params["access_key_id"] = body.pop("access_key_id")
            params["secret_access_key"] = body.pop("secret_access_key")
            if body.get("new_access_key_id"):
                body["access_key_id"] = body.pop("new_access_key_id")
            if body.get("new_secret_access_key"):
                body["secret_access_key"] = body.pop("new_secret_access_key")
        return super(AwsIntegration, cls).update(id=cls._resource_id, params=params, **body)

    @classmethod
    def delete(cls, **body):
        """
        Delete a given Datadog-AWS integration.

        >>> account_id = "<AWS_ACCOUNT_ID>"
        >>> role_name = "<Datadog Integration Role Name>"

        >>> api.AwsIntegration.delete()
        """
        return super(AwsIntegration, cls).delete(id=cls._resource_id, body=body)

    @classmethod
    def list_namespace_rules(cls, **params):
        """
        List all namespace rules available as options.

        >>> api.AwsIntegration.list_namespace_rules()
        """
        cls._sub_resource_name = "available_namespace_rules"
        return super(AwsIntegration, cls).get_items(id=cls._resource_id, **params)

    @classmethod
    def generate_new_external_id(cls, **params):
        """
        Generate a new AWS external id for a given AWS account id and role name pair.

        >>> account_id = "<AWS_ACCOUNT_ID>"
        >>> role_name = "<Datadog Integration Role Name>"

        >>> api.AwsIntegration.generate_new_external_id()
        """
        cls._sub_resource_name = "generate_new_external_id"
        return super(AwsIntegration, cls).update_items(id=cls._resource_id, **params)



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
# stdlib
import argparse
import json

# 3p
from datadog.util.cli import (
    set_of_ints,
    comma_set,
    comma_list_or_empty,
    parse_date_as_epoch_timestamp,
)
from datadog.util.format import pretty_json

# datadog
from datadog import api
from datadog.dogshell.common import report_errors, report_warnings


class ServiceLevelObjectiveClient(object):
    @classmethod
    def setup_parser(cls, subparsers):
        parser = subparsers.add_parser(
            "service_level_objective",
            help="Create, edit, and delete service level objectives",
        )

        verb_parsers = parser.add_subparsers(title="Verbs", dest="verb")
        verb_parsers.required = True

        create_parser = verb_parsers.add_parser("create", help="Create a SLO")
        create_parser.add_argument(
            "--type",
            help="type of the SLO, e.g.",
            choices=["metric", "monitor"],
        )
        create_parser.add_argument("--name", help="name of the SLO", default=None)
        create_parser.add_argument("--description", help="description of the SLO", default=None)
        create_parser.add_argument(
            "--tags",
            help="comma-separated list of tags",
            default=None,
            type=comma_list_or_empty,
        )
        create_parser.add_argument(
            "--thresholds",
            help="comma separated list of <timeframe>:<target>[:<warning>[:<target_display>[:<warning_display>]]",
        )
        create_parser.add_argument(
            "--numerator",
            help="numerator metric query (sum of good events)",
            default=None,
        )
        create_parser.add_argument(
            "--denominator",
            help="denominator metric query (sum of total events)",
            default=None,
        )
        create_parser.add_argument(
            "--monitor_ids",
            help="explicit monitor_ids to use (CSV)",
            default=None,
            type=set_of_ints,
        )
        create_parser.add_argument("--monitor_search", help="monitor search terms to use", default=None)
        create_parser.add_argument(
            "--groups",
            help="for a single monitor you can specify the specific groups as a pipe (|) delimited string",
            default=None,
            type=comma_list_or_empty,
        )
        create_parser.set_defaults(func=cls._create)

        file_create_parser = verb_parsers.add_parser("fcreate", help="Create a SLO from file")
        file_create_parser.add_argument("file", help="json file holding all details", type=argparse.FileType("r"))
        file_create_parser.set_defaults(func=cls._file_create)

        update_parser = verb_parsers.add_parser("update", help="Update existing SLO")
        update_parser.add_argument("slo_id", help="SLO to replace with the new definition")
        update_parser.add_argument(
            "--type",
            help="type of the SLO (must specify it's original type)",
            choices=["metric", "monitor"],
        )
        update_parser.add_argument("--name", help="name of the SLO", default=None)
        update_parser.add_argument("--description", help="description of the SLO", default=None)
        update_parser.add_argument(
            "--thresholds",
            help="comma separated list of <timeframe>:<target>[:<warning>[:<target_display>[:<warning_display>]]",
        )
        update_parser.add_argument(
            "--tags",
            help="comma-separated list of tags",
            default=None,
            type=comma_list_or_empty,
        )
        update_parser.add_argument(
            "--numerator",
            help="numerator metric query (sum of good events)",
            default=None,
        )
        update_parser.add_argument(
            "--denominator",
            help="denominator metric query (sum of total events)",
            default=None,
        )
        update_parser.add_argument(
            "--monitor_ids",
            help="explicit monitor_ids to use (CSV)",
            default=[],
            type=list,
        )
        update_parser.add_argument("--monitor_search", help="monitor search terms to use", default=None)
        update_parser.add_argument(
            "--groups",
            help="for a single monitor you can specify the specific groups as a pipe (|) delimited string",
            default=None,
        )
        update_parser.set_defaults(func=cls._update)

        file_update_parser = verb_parsers.add_parser("fupdate", help="Update existing SLO from file")
        file_update_parser.add_argument("file", help="json file holding all details", type=argparse.FileType("r"))
        file_update_parser.set_defaults(func=cls._file_update)

        show_parser = verb_parsers.add_parser("show", help="Show a SLO definition")
        show_parser.add_argument("slo_id", help="SLO to show")
        show_parser.set_defaults(func=cls._show)

        show_all_parser = verb_parsers.add_parser("show_all", help="Show a list of all SLOs")
        show_all_parser.add_argument("--query", help="string to filter SLOs by query (see UI or documentation)")
        show_all_parser.add_argument(
            "--slo_ids",
            help="comma separated list indicating what SLO IDs to get at once",
            type=comma_set,
        )
        show_all_parser.add_argument("--offset", help="offset of query pagination", default=0)
        show_all_parser.add_argument("--limit", help="limit of query pagination", default=100)
        show_all_parser.set_defaults(func=cls._show_all)

        delete_parser = verb_parsers.add_parser("delete", help="Delete a SLO")
        delete_parser.add_argument("slo_id", help="SLO to delete")
        delete_parser.set_defaults(func=cls._delete)

        delete_many_parser = verb_parsers.add_parser("delete_many", help="Delete a SLO")
        delete_many_parser.add_argument("slo_ids", help="comma separated list of SLO IDs to delete", type=comma_set)
        delete_many_parser.set_defaults(func=cls._delete_many)

        delete_timeframe_parser = verb_parsers.add_parser("delete_many", help="Delete a SLO timeframe")
        delete_timeframe_parser.add_argument("slo_id", help="SLO ID to update")
        delete_timeframe_parser.add_argument(
            "timeframes",
            help="CSV of timeframes to delete, e.g. 7d,30d,90d",
            type=comma_set,
        )
        delete_timeframe_parser.set_defaults(func=cls._delete_timeframe)

        can_delete_parser = verb_parsers.add_parser("can_delete", help="Check if can delete SLOs")
        can_delete_parser.add_argument("slo_ids", help="comma separated list of SLO IDs to delete", type=comma_set)
        can_delete_parser.set_defaults(func=cls._can_delete)

        history_parser = verb_parsers.add_parser("history", help="Get the SLO history")
        history_parser.add_argument("slo_id", help="SLO to query the history")
        history_parser.add_argument(
            "from_ts",
            type=parse_date_as_epoch_timestamp,
            help="`from` date or timestamp",
        )
        history_parser.add_argument(
            "to_ts",
            type=parse_date_as_epoch_timestamp,
            help="`to` date or timestamp",
        )
        history_parser.set_defaults(func=cls._history)

    @classmethod
    def _create(cls, args):
        api._timeout = args.timeout
        format = args.format

        params = {"type": args.type, "name": args.name}

        if args.tags:
            tags = sorted(set([t.strip() for t in args.tags.split(",") if t.strip()]))
            params["tags"] = tags

        thresholds = []
        for threshold_str in args.thresholds.split(","):
            parts = threshold_str.split(":")
            timeframe = parts[0]
            target = float(parts[1])

            threshold = {"timeframe": timeframe, "target": target}

            if len(parts) > 2:
                threshold["warning"] = float(parts[2])

            if len(parts) > 3 and parts[3]:
                threshold["target_display"] = parts[3]

            if len(parts) > 4 and parts[4]:
                threshold["warning_display"] = parts[4]

            thresholds.append(threshold)
        params["thresholds"] = thresholds

        if args.description:
            params["description"] = args.description

        if args.type == "metric":
            params["query"] = {
                "numerator": args.numerator,
                "denominator": args.denominator,
            }
        elif args.monitor_search:
            params["monitor_search"] = args.monitor_search
        else:
            params["monitor_ids"] = list(args.monitor_ids)
            if args.groups and len(args.monitor_ids) == 1:
                groups = args.groups.split("|")
                params["groups"] = groups

        if args.tags:
            params["tags"] = args.tags

        res = api.ServiceLevelObjective.create(**params)
        report_warnings(res)
        report_errors(res)
        if format == "pretty":
            print(pretty_json(res))
        else:
            print(json.dumps(res))

    @classmethod
    def _file_create(cls, args):
        api._timeout = args.timeout
        format = args.format
        slo = json.load(args.file)
        res = api.ServiceLevelObjective.create(return_raw=True, **slo)
        report_warnings(res)
        report_errors(res)
        if format == "pretty":
            print(pretty_json(res))
        else:
            print(json.dumps(res))

    @classmethod
    def _update(cls, args):
        api._timeout = args.timeout
        format = args.format

        params = {"type": args.type}

        if args.thresholds:
            thresholds = []
            for threshold_str in args.thresholds.split(","):
                parts = threshold_str.split(":")
                timeframe = parts[0]
                target = parts[1]

                threshold = {"timeframe": timeframe, "target": target}

                if len(parts) > 2:
                    threshold["warning"] = float(parts[2])

                if len(parts) > 3 and parts[3]:
                    threshold["target_display"] = parts[3]

                if len(parts) > 4 and parts[4]:
                    threshold["warning_display"] = parts[4]

                thresholds.append(threshold)
            params["thresholds"] = thresholds

        if args.description:
            params["description"] = args.description

        if args.type == "metric":
            if args.numerator and args.denominator:
                params["query"] = {
                    "numerator": args.numerator,
                    "denominator": args.denominator,
                }
        elif args.monitor_search:
            params["monitor_search"] = args.monitor_search
        else:
            params["monitor_ids"] = args.monitor_ids
            if args.groups and len(args.monitor_ids) == 1:
                groups = args.groups.split("|")
                params["groups"] = groups

        if args.tags:
            tags = sorted(set([t.strip() for t in args.tags if t.strip()]))
            params["tags"] = tags
        res = api.ServiceLevelObjective.update(args.slo_id, return_raw=True, **params)
        report_warnings(res)
        report_errors(res)
        if format == "pretty":
            print(pretty_json(res))
        else:
            print(json.dumps(res))

    @classmethod
    def _file_update(cls, args):
        api._timeout = args.timeout
        format = args.format
        slo = json.load(args.file)

        res = api.ServiceLevelObjective.update(slo["id"], return_raw=True, **slo)
        report_warnings(res)
        report_errors(res)
        if format == "pretty":
            print(pretty_json(res))
        else:
            print(json.dumps(res))

    @classmethod
    def _show(cls, args):
        api._timeout = args.timeout
        format = args.format
        res = api.ServiceLevelObjective.get(args.slo_id, return_raw=True)
        report_warnings(res)
        report_errors(res)

        if args.string_ids:
            res["id"] = str(res["id"])

        if format == "pretty":
            print(pretty_json(res))
        else:
            print(json.dumps(res))

    @classmethod
    def _show_all(cls, args):
        api._timeout = args.timeout
        format = args.format

        params = {"offset": args.offset, "limit": args.limit}
        if args.query:
            params["query"] = args.query
        else:
            params["ids"] = args.slo_ids

        res = api.ServiceLevelObjective.get_all(return_raw=True, **params)
        report_warnings(res)
        report_errors(res)

        if format == "pretty":
            print(pretty_json(res))
        else:
            print(json.dumps(res))

    @classmethod
    def _delete(cls, args):
        api._timeout = args.timeout
        res = api.ServiceLevelObjective.delete(args.slo_id, return_raw=True)
        if res is not None:
            report_warnings(res)
            report_errors(res)

            if format == "pretty":
                print(pretty_json(res))
            else:
                print(json.dumps(res))

    @classmethod
    def _delete_many(cls, args):
        api._timeout = args.timeout
        res = api.ServiceLevelObjective.delete_many(args.slo_ids)
        if res is not None:
            report_warnings(res)
            report_errors(res)

            if format == "pretty":
                print(pretty_json(res))
            else:
                print(json.dumps(res))

    @classmethod
    def _delete_timeframe(cls, args):
        api._timeout = args.timeout

        ops = {args.slo_id: args.timeframes}

        res = api.ServiceLevelObjective.bulk_delete(ops)
        if res is not None:
            report_warnings(res)
            report_errors(res)

            if format == "pretty":
                print(pretty_json(res))
            else:
                print(json.dumps(res))

    @classmethod
    def _can_delete(cls, args):
        api._timeout = args.timeout

        res = api.ServiceLevelObjective.can_delete(args.slo_ids)
        if res is not None:
            report_warnings(res)
            report_errors(res)

            if format == "pretty":
                print(pretty_json(res))
            else:
                print(json.dumps(res))

    @classmethod
    def _history(cls, args):
        api._timeout = args.timeout

        res = api.ServiceLevelObjective.history(args.slo_id)
        if res is not None:
            report_warnings(res)
            report_errors(res)

            if format == "pretty":
                print(pretty_json(res))
            else:
                print(json.dumps(res))

    @classmethod
    def _escape(cls, s):
        return s.replace("\r", "\\r").replace("\n", "\\n").replace("\t", "\\t")



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
import os
import string
import sys

# datadog
from datadog.util.compat import configparser, StringIO, is_p3k
from datadog.version import __version__

# CONSTANTS
DATADOG_CONF = "datadog.conf"


class CfgNotFound(Exception):
    pass


class PathNotFound(Exception):
    pass


def get_os():
    "Human-friendly OS name"
    if sys.platform == "darwin":
        return "mac"
    elif sys.platform.find("freebsd") != -1:
        return "freebsd"
    elif sys.platform.find("linux") != -1:
        return "linux"
    elif sys.platform.find("win32") != -1:
        return "windows"
    elif sys.platform.find("sunos") != -1:
        return "solaris"
    else:
        return sys.platform


def skip_leading_wsp(f):
    "Works on a file, returns a file-like object"
    if is_p3k():
        return StringIO("\n".join(x.strip(" ") for x in f.readlines()))
    else:
        return StringIO("\n".join(map(string.strip, f.readlines())))


def _windows_commondata_path():
    """Return the common appdata path, using ctypes
    From http://stackoverflow.com/questions/626796/\
    how-do-i-find-the-windows-common-application-data-folder-using-python
    """
    import ctypes
    from ctypes import wintypes, windll

    CSIDL_COMMON_APPDATA = 35

    _SHGetFolderPath = windll.shell32.SHGetFolderPathW
    _SHGetFolderPath.argtypes = [wintypes.HWND, ctypes.c_int, wintypes.HANDLE, wintypes.DWORD, wintypes.LPCWSTR]

    path_buf = ctypes.create_unicode_buffer(wintypes.MAX_PATH)
    _SHGetFolderPath(0, CSIDL_COMMON_APPDATA, 0, 0, path_buf)
    return path_buf.value


def _windows_config_path():
    common_data = _windows_commondata_path()
    path = os.path.join(common_data, "Datadog", DATADOG_CONF)
    if os.path.exists(path):
        return path
    raise PathNotFound(path)


def _unix_config_path():
    path = os.path.join("/etc/dd-agent", DATADOG_CONF)
    if os.path.exists(path):
        return path
    raise PathNotFound(path)


def _mac_config_path():
    path = os.path.join("~/.datadog-agent/agent", DATADOG_CONF)
    path = os.path.expanduser(path)
    if os.path.exists(path):
        return path
    raise PathNotFound(path)


def get_config_path(cfg_path=None, os_name=None):
    # Check if there's an override and if it exists
    if cfg_path is not None and os.path.exists(cfg_path):
        return cfg_path

    if os_name is None:
        os_name = get_os()

    # Check for an OS-specific path, continue on not-found exceptions
    if os_name == "windows":
        return _windows_config_path()
    elif os_name == "mac":
        return _mac_config_path()
    else:
        return _unix_config_path()


def get_config(cfg_path=None, options=None):
    agentConfig = {}

    # Config handling
    try:
        # Find the right config file
        path = os.path.realpath(__file__)
        path = os.path.dirname(path)

        config_path = get_config_path(cfg_path, os_name=get_os())
        config = configparser.ConfigParser()
        with open(config_path) as config_file:
            if is_p3k():
                config.read_file(skip_leading_wsp(config_file))
            else:
                config.readfp(skip_leading_wsp(config_file))

        # bulk import
        for option in config.options("Main"):
            agentConfig[option] = config.get("Main", option)

    except Exception:
        raise CfgNotFound

    return agentConfig


def get_pkg_version():
    """
    Resolve `datadog` package version.

    Deprecated: use `datadog.__version__` directly instead
    """
    return __version__


def get_version():
    """
    Resolve `datadog` package version.

    Deprecated: use `datadog.__version__` directly instead
    """
    return __version__



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
# stdlib
import json

# 3p
from datadog.util.format import pretty_json

# datadog
from datadog import api
from datadog.dogshell.common import report_errors, report_warnings


class ServiceCheckClient(object):
    @classmethod
    def setup_parser(cls, subparsers):
        parser = subparsers.add_parser("service_check", help="Perform service checks")
        verb_parsers = parser.add_subparsers(title="Verbs", dest="verb")
        verb_parsers.required = True

        check_parser = verb_parsers.add_parser("check", help="text for the message")
        check_parser.add_argument("check", help="text for the message")
        check_parser.add_argument("host_name", help="name of the host submitting the check")
        check_parser.add_argument(
            "status",
            help="integer for the status of the check." " i.e: '0': OK, '1': WARNING, '2': CRITICAL, '3': UNKNOWN",
        )
        check_parser.add_argument("--timestamp", help="POSIX timestamp of the event", default=None)
        check_parser.add_argument("--message", help="description of why this status occurred", default=None)
        check_parser.add_argument("--tags", help="comma separated list of tags", default=None)
        check_parser.set_defaults(func=cls._check)

    @classmethod
    def _check(cls, args):
        api._timeout = args.timeout
        format = args.format
        if args.tags:
            tags = sorted(set([t.strip() for t in args.tags.split(",") if t.strip()]))
        else:
            tags = None
        res = api.ServiceCheck.check(
            check=args.check,
            host_name=args.host_name,
            status=int(args.status),
            timestamp=args.timestamp,
            message=args.message,
            tags=tags,
        )
        report_warnings(res)
        report_errors(res)
        if format == "pretty":
            print(pretty_json(res))
        else:
            print(json.dumps(res))



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
from datadog.api.resources import GetableAPIResource, CreateableAPIResource, DeletableAPIResource, UpdatableAPIResource


class GcpIntegration(GetableAPIResource, CreateableAPIResource, DeletableAPIResource, UpdatableAPIResource):
    """
    A wrapper around GCP integration API.
    """

    _resource_name = "integration"
    _resource_id = "gcp"

    @classmethod
    def list(cls, **params):
        """
        List all Datadog-Gcp integrations available in your Datadog organization.

        >>> api.GcpIntegration.list()
        """
        return super(GcpIntegration, cls).get(id=cls._resource_id, **params)

    @classmethod
    def delete(cls, **body):
        """
        Delete a given Datadog-GCP integration.

        >>> project_id="<GCP_CLIENT_ID>"
        >>> client_email="<GCP_CLIENT_EMAIL>"

        >>> api.GcpIntegration.delete(project_id=project_id, client_email=client_email)
        """
        return super(GcpIntegration, cls).delete(id=cls._resource_id, body=body)

    @classmethod
    def create(cls, **params):
        """
        Add a new GCP integration config.

        All of the following fields values are provided by the \
        JSON service account key file created in the GCP Console \
        for service accounts; Refer to the Datadog-Google Cloud \
        Platform integration installation instructions to see how \
        to generate one for your organization. For further references, \
        consult the Google Cloud service account documentation.

        >>> type="service_account"
        >>> project_id="<GCP_PROJECT_ID>"
        >>> private_key_id="<GCP_PRIVATE_KEY_ID>"
        >>> private_key="<GCP_PRIVATE_KEY>"
        >>> client_email="<GCP_CLIENT_EMAIL>"
        >>> client_id="<GCP_CLIENT_ID>"
        >>> auth_uri="<GCP_AUTH_URI"
        >>> token_uri="<GCP_TOKEN_URI>"
        >>> auth_provider_x509_cert_url="<GCP_AUTH_PROVIDER_X509_CERT_URL>"
        >>> client_x509_cert_url="<GCP_CLIENT_X509_CERT_URL>"
        >>> host_filters="<KEY>:<VALUE>,<KEY>:<VALUE>"

        >>> api.GcpIntegration.create(type=type, project_id=project_id, \
        private_key_id=private_key_id,private_key=private_key, \
        client_email=client_email, client_id=client_id, \
        auth_uri=auth_uri, token_uri=token_uri, \
        auth_provider_x509_cert_url=auth_provider_x509_cert_url, \
        client_x509_cert_url=client_x509_cert_url, host_filters=host_filters)
        """
        return super(GcpIntegration, cls).create(id=cls._resource_id, **params)

    @classmethod
    def update(cls, **body):
        """
        Update an existing service account partially (one or multiple fields), \
        by supplying a new value for the field(s) to be updated.

        `project_id` and `client_email` are required, in order to identify the \
        right service account to update. \
        The unspecified fields will keep their original values.

        The only use case for updating this integration is to change \
        host filtering and automute settings. Otherwise, an entirely \
        new integration config is needed.

        >>> project_id="<GCP_PROJECT_ID>"
        >>> client_email="<GCP_CLIENT_EMAIL>"
        >>> host_filters="<NEW_HOST_FILTERS>"
        >>> automute=true #boolean

        >>> api.GcpIntegration.update(project_id=project_id, \
        client_email=client_email, host_filters=host_filters, \
        automute=automute)
        """
        params = {}
        return super(GcpIntegration, cls).update(id=cls._resource_id, params=params, **body)
